"""Eclass parsing support."""

import os
import re
import shlex
import subprocess
from datetime import datetime
from functools import partial

from snakeoil import klass
from snakeoil.mappings import ImmutableDict, OrderedSet
from snakeoil.strings import pluralism
from snakeoil.version import get_version

from .. import __title__
from ..log import logger
from . import conditionals
from .eapi import EAPI


class AttrDict(ImmutableDict):
    """Support accessing dict keys as attributes."""

    def __getattr__(self, name):
        try:
            object.__getattribute__(self, name)
        except AttributeError as e:
            try:
                return object.__getattribute__(self, '_dict')[name]
            except KeyError:
                raise e

    def __dir__(self):
        return sorted(dir(self._dict) + list(self._dict))


def _rst_header(char, text, leading=False, newline=False):
    """Create rST header data from a given character and header text."""
    sep = char * len(text)
    data = [text, sep]
    if leading:
        data = [sep] + data
    if newline:
        data.append('')
    return data


class ParseEclassDoc:
    """Generic block for eclass docs.

    See the devmanual [#]_ for the eclass docs specification.

    .. [#] https://devmanual.gentoo.org/eclass-writing/#documenting-eclasses
    """

    # block tag
    tag = None
    # block key name -- None for singular block types
    key = None
    # boolean flagging if block attribute defaults should be inserted
    default = False

    # mapping from eclass doc blocks to parsing instances
    blocks = {}

    def __init_subclass__(cls, **kwargs):
        """Register known eclass doc block tags."""
        super().__init_subclass__(**kwargs)
        cls.blocks[cls.tag] = cls()

    def __init__(self, tags):
        self.tags = tags
        # regex matching all known tags for the eclass doc block
        self._block_tags_re = re.compile(rf'^(?P<tag>{"|".join(self.tags)})(?P<value>.*)')
        # regex matching @CODE tags
        self._code_tag = re.compile(r'^\s*@CODE\s*$')
        # regex matching @SUBSECTION tags
        self._subsection_tag = re.compile(r'^\s*@SUBSECTION (?P<title>.+)$')

    def _tag_bool(self, block, tag, lineno):
        """Parse boolean tags."""
        try:
            args = next(x for x in block if x)
            logger.warning(
                f'{repr(tag)}, line {lineno}: tag takes no args, got {repr(args)}')
        except StopIteration:
            pass
        return True

    def _tag_inline_arg(self, block, tag, lineno):
        """Parse tags with inline argument."""
        if not block[0]:
            logger.warning(f'{repr(tag)}, line {lineno}: missing inline arg')
        elif len(block) > 1:
            logger.warning(f'{repr(tag)}, line {lineno}: non-inline arg')
        return block[0]

    def _tag_inline_list(self, block, tag, lineno):
        """Parse tags with inline, space-separated list argument."""
        line = self._tag_inline_arg(block, tag, lineno)
        return tuple(line.split())

    def _tag_multiline_args(self, block, tag, lineno):
        """Parse tags with multiline arguments."""
        if block[0]:
            logger.warning(f'{repr(tag)}, line {lineno}: invalid inline arg')
        if not block[1:]:
            logger.warning(f'{repr(tag)}, line {lineno}: missing args')
        return tuple(block[1:])

    def _tag_multiline_str(self, block, tag, lineno):
        """Parse tags with multiline text while handling @CODE/@SUBSECTION tags."""
        lines = self._tag_multiline_args(block, tag, lineno)
        if not lines:
            return None

        # use literal blocks for all multiline text
        data = ['::', '\n\n']

        for i, line in enumerate(lines, 1):
            if self._code_tag.match(line):
                continue
            elif mo := self._subsection_tag.match(line):
                header = _rst_header('~', mo.group('title'))
                data.extend(f'{x}\n' for x in header)
                data.extend(['::', '\n\n'])
            elif line:
                data.append(f'  {line}\n')
            else:
                data.append('\n')

        return ''.join(data).rstrip('\n')

    def _tag_multiline_rst(self, block, tag, lineno):
        """Parse tags with multiline rST formatting."""
        lines = self._tag_multiline_args(block, tag, lineno)
        return ''.join(lines).rstrip('\n')

    def _tag_deprecated(self, block, tag, lineno):
        """Parse deprecated tags."""
        arg = self._tag_inline_arg(block, tag, lineno)
        return True if arg.lower() == 'none' else arg

    @klass.jit_attr
    def _required(self):
        """Set of required eclass doc block tags."""
        tags = set()
        for tag, (_name, required, _func, _default) in self.tags.items():
            if required:
                tags.add(tag)
        return frozenset(tags)

    @klass.jit_attr
    def bash_env_vars(self):
        """The set of all bash variables defined in the default environment."""
        variables = []
        # use no-op to fake a pipeline so pipeline specific vars are defined
        p = subprocess.run(
            ['bash', '-c', ':; compgen -A variable'],
            stderr=subprocess.DEVNULL, stdout=subprocess.PIPE, encoding='utf8')
        if p.returncode == 0:
            variables = p.stdout.splitlines()
        return frozenset(variables)

    @property
    def defaults(self):
        """Return default field mapping for the block."""
        return {name: default for name, _required, _func, default in self.tags.values()}

    def parse(self, lines, line_ind):
        """Parse an eclass block."""
        blocks = []
        data = self.defaults
        # track if all required tags are defined
        missing_tags = set(self._required)

        # split eclass doc block into separate blocks by tag
        for i, line in enumerate(lines):
            if (mo := self._block_tags_re.match(line)):
                tag = mo.group('tag')
                missing_tags.discard(tag)
                value = mo.group('value').strip()
                blocks.append((tag, line_ind + i, [value]))
            else:
                blocks[-1][-1].append(line)

        # parse each tag block
        for tag, line_ind, block in blocks:
            name, required, func, _default = self.tags[tag]
            data[name] = func(block, tag, line_ind)

        # check if any required tags are missing
        if missing_tags:
            missing_tags_str = ', '.join(map(repr, missing_tags))
            s = pluralism(missing_tags)
            logger.warning(
                f'{repr(lines[0])}: missing tag{s}: {missing_tags_str}')

        return AttrDict(data)


class EclassBlock(ParseEclassDoc):
    """ECLASS doc block."""

    tag = '@ECLASS:'

    def __init__(self):
        tags = {
            '@ECLASS:': ('name', True, self._tag_inline_arg, None),
            '@VCSURL:': ('vcsurl', False, self._tag_inline_arg, None),
            '@BLURB:': ('blurb', True, self._tag_inline_arg, None),
            '@DEPRECATED:': ('deprecated', False, self._tag_deprecated, False),
            '@PROVIDES:': ('raw_provides', False, self._tag_inline_list, ()),
            '@MAINTAINER:': ('maintainers', True, self._tag_multiline_args, None),
            '@AUTHOR:': ('authors', False, self._tag_multiline_args, None),
            '@BUGREPORTS:': ('bugreports', False, self._tag_multiline_str, None),
            '@DESCRIPTION:': ('description', False, self._tag_multiline_str, None),
            '@EXAMPLE:': ('example', False, self._tag_multiline_str, None),
            '@SUPPORTED_EAPIS:': ('supported_eapis', False, self._supported_eapis, ()),
        }
        super().__init__(tags)

        self._known_eapis = frozenset(EAPI.known_eapis)

    def _supported_eapis(self, block, tag, lineno):
        """Parse @SUPPORTED_EAPIS tag arguments."""
        eapis = self._tag_inline_list(block, tag, lineno)
        unknown = set(eapis) - self._known_eapis
        if unknown:
            s = pluralism(unknown)
            unknown_str = ' '.join(sorted(unknown))
            logger.warning(
                f'{repr(tag)}, line {lineno}: unknown EAPI{s}: {unknown_str}')
        return OrderedSet(eapis)


class EclassVarBlock(ParseEclassDoc):
    """ECLASS_VARIABLE doc block."""

    tag = '@ECLASS_VARIABLE:'
    key = 'variables'
    default = True

    def __init__(self):
        tags = {
            '@ECLASS_VARIABLE:': ('name', True, self._tag_inline_arg, None),
            '@DEPRECATED:': ('deprecated', False, self._tag_deprecated, False),
            '@DEFAULT_UNSET': ('default_unset', False, self._tag_bool, False),
            '@INTERNAL': ('internal', False, self._tag_bool, False),
            '@REQUIRED': ('required', False, self._tag_bool, False),
            '@PRE_INHERIT': ('pre_inherit', False, self._tag_bool, False),
            '@USER_VARIABLE': ('user_variable', False, self._tag_bool, False),
            '@OUTPUT_VARIABLE': ('output_variable', False, self._tag_bool, False),
            '@DESCRIPTION:': ('description', True, self._tag_multiline_str, None),
        }
        super().__init__(tags)


# For backwards compatibility, can be removed after 2022-12-31
class EclassVarBlockCompat(ParseEclassDoc):
    """ECLASS-VARIABLE doc block."""

    tag = '@ECLASS-VARIABLE:'
    key = 'variables'
    default = True

    def __init__(self):
        tags = {
            '@ECLASS-VARIABLE:': ('name', True, self._eclass_variable, None),
            '@DEPRECATED:': ('deprecated', False, self._tag_deprecated, False),
            '@DEFAULT_UNSET': ('default_unset', False, self._tag_bool, False),
            '@INTERNAL': ('internal', False, self._tag_bool, False),
            '@REQUIRED': ('required', False, self._tag_bool, False),
            '@PRE_INHERIT': ('pre_inherit', False, self._tag_bool, False),
            '@USER_VARIABLE': ('user_variable', False, self._tag_bool, False),
            '@OUTPUT_VARIABLE': ('output_variable', False, self._tag_bool, False),
            '@DESCRIPTION:': ('description', True, self._tag_multiline_str, None),
        }
        super().__init__(tags)

    def _eclass_variable(self, block, tag, lineno):
        """Parse @ECLASS-VARIABLE tag."""
        logger.warning(f"{repr(tag)}, line {lineno}: deprecated, use '@ECLASS_VARIABLE' instead")
        return self._tag_inline_arg(block, tag, lineno)


class EclassFuncBlock(ParseEclassDoc):
    """FUNCTION doc block."""

    tag = '@FUNCTION:'
    key = 'functions'
    default = True

    def __init__(self):
        tags = {
            '@FUNCTION:': ('name', True, self._tag_inline_arg, None),
            '@RETURN:': ('returns', False, self._tag_inline_arg, None),
            '@DEPRECATED:': ('deprecated', False, self._tag_deprecated, False),
            '@INTERNAL': ('internal', False, self._tag_bool, False),
            '@MAINTAINER:': ('maintainers', False, self._tag_multiline_args, None),
            '@DESCRIPTION:': ('description', False, self._tag_multiline_str, None),
            # TODO: The devmanual states this is required, but disabling for now since
            # many phase override functions don't document usage.
            '@USAGE:': ('usage', False, self._usage, None),
        }
        super().__init__(tags)

    def _usage(self, block, tag, lineno):
        """Parse @USAGE tag arguments.

        Empty usage is allowed for functions with no arguments.
        """
        if len(block) > 1:
            logger.warning(f'{repr(tag)}, line {lineno}: non-inline arg')
        return block[0]

    def parse(self, *args):
        data = super().parse(*args)
        if not (data.returns or data.description):
            logger.warning(f"'{self.tag}:{data.name}', @RETURN or @DESCRIPTION required")
        return data


class EclassFuncVarBlock(ParseEclassDoc):
    """VARIABLE doc block."""

    tag = '@VARIABLE:'
    key = 'function_variables'
    default = True

    def __init__(self):
        tags = {
            '@VARIABLE:': ('name', True, self._tag_inline_arg, None),
            '@DEPRECATED:': ('deprecated', False, self._tag_deprecated, False),
            '@DEFAULT_UNSET': ('default_unset', False, self._tag_bool, False),
            '@INTERNAL': ('internal', False, self._tag_bool, False),
            '@REQUIRED': ('required', False, self._tag_bool, False),
            '@DESCRIPTION:': ('description', True, self._tag_multiline_str, None),
        }
        super().__init__(tags)


_eclass_blocks_re = re.compile(
    rf'^(?P<prefix>\s*#) (?P<tag>{"|".join(ParseEclassDoc.blocks)})(?P<value>.*)')


class EclassDoc(AttrDict):
    """Support parsing eclass docs for a given eclass path."""

    ABI_VERSION = 5

    def __init__(self, path, /, *, sourced=False, repo=None):
        self.mtime = os.path.getmtime(path)

        # parse eclass doc
        data = self.parse(path)

        data['provides'] = None
        if repo is not None:
            data['provides'] = self._get_provides(data['raw_provides'], repo)

        # inject full lists of exported funcs and vars
        if sourced:
            data.update(self._source_eclass(path))

        super().__init__(data)

    @staticmethod
    def _get_provides(raw_provides, repo):
        out = OrderedSet()
        to_process = OrderedSet(raw_provides)
        while to_process:
            out.add(next_eclass := to_process.pop())
            if (next_eclass_inst := repo.eclass_cache.get_eclass(next_eclass)) is None:
                logger.warning(f"'@PROVIDES:' eclass {next_eclass} not found")
                continue
            next_eclass_doc = EclassDoc(next_eclass_inst.path)
            to_process.update(set(next_eclass_doc.raw_provides).difference(out))
        return tuple(out)

    @staticmethod
    def _source_eclass(path):
        data = {}
        # TODO: support this via pkgcore's ebd
        # source eclass to determine PROPERTIES
        p = subprocess.run(
            ['bash', '-c',
                f'source {shlex.quote(path)}; '
                f'compgen -A function; '
                f'echo "#"; '
                f'compgen -A variable; '
                f'echo "#"; '
                f'echo ${{PROPERTIES}}'],
            stderr=subprocess.DEVNULL, stdout=subprocess.PIPE, encoding='utf8')
        if p.returncode == 0:
            eclass_obj = ParseEclassDoc.blocks['@ECLASS:']
            funcs, variables, properties = p.stdout.split('#\n')
            data['_exported_funcs'] = tuple(funcs.split())
            data['_exported_vars'] = tuple(
                x for x in variables.split()
                if x not in eclass_obj.bash_env_vars
            )
            data['_properties'] = conditionals.DepSet.parse(
                properties, str, operators={}, attr='PROPERTIES')
        return data

    @property
    def function_names(self):
        """Set of documented function names in the eclass."""
        return frozenset(x.name for x in self.functions)

    @property
    def internal_function_names(self):
        """Set of internal function names in the eclass."""
        # include all internal tagged functions
        s = {x.name for x in self.functions if x.internal}
        # and all exported, underscore-prefixed functions
        s.update(x for x in self._dict.get('_exported_funcs', ()) if x.startswith('_'))
        return frozenset(s)

    @property
    def exported_function_names(self):
        """Set of all exported function names in the eclass."""
        return frozenset(self._dict.get('_exported_funcs', ()))

    @property
    def variable_names(self):
        """Set of documented variable names in the eclass."""
        return frozenset(x.name for x in self.variables)

    @property
    def internal_variable_names(self):
        """Set of internal variable names in the eclass."""
        # include all internal tagged variables
        s = {x.name for x in self.variables if x.internal}
        # and all exported, underscore-prefixed variables
        s.update(x for x in self._dict.get('_exported_vars', ()) if x.startswith('_'))
        return frozenset(s)

    @property
    def exported_variable_names(self):
        """Set of all exported variable names in the eclass.

        Ignores variables that start with underscores since
        it's assumed they are private.
        """
        return frozenset(self._dict.get('_exported_vars', ()))

    @property
    def function_variable_names(self):
        """Set of documented function variable names in the eclass."""
        return frozenset(x.name for x in self.function_variables)

    @property
    def live(self):
        """Eclass implements functionality to support a version control system."""
        return 'live' in self._dict.get('_properties', ())

    @staticmethod
    def parse(path):
        """Parse eclass docs."""
        blocks = []

        with open(path) as f:
            lines = f.read().splitlines()
            line_ind = 0

            while line_ind < len(lines):
                if (mo := _eclass_blocks_re.match(lines[line_ind])):
                    # Isolate identified doc block by pulling all following
                    # lines with a matching prefix.
                    prefix = mo.group('prefix')
                    tag = mo.group('tag')
                    block = []
                    block_start = line_ind + 1
                    while line_ind < len(lines):
                        line = lines[line_ind]
                        if not line.startswith(prefix):
                            break
                        line = line[len(prefix) + 1:]
                        block.append(line)
                        line_ind += 1
                    blocks.append((tag, block, block_start))
                line_ind += 1

        # set default fields
        data = {}
        data.update(ParseEclassDoc.blocks['@ECLASS:'].defaults)
        for block_obj in ParseEclassDoc.blocks.values():
            if block_obj.default:
                data[block_obj.key] = OrderedSet()
        data.update({block.key: OrderedSet() for block in ParseEclassDoc.blocks.values() if block.default})

        # @ECLASS block must exist and be first in eclasses
        if not blocks:
            logger.error("'@ECLASS:' block missing")
            return data
        elif blocks[0][0] != '@ECLASS:':
            logger.warning("'@ECLASS:' block not first")

        # track duplicate tags
        duplicates = {k: set() for k in ParseEclassDoc.blocks}

        # parse identified blocks
        for tag, block, block_start in blocks:
            block_obj = ParseEclassDoc.blocks[tag]
            block_data = block_obj.parse(block, block_start)
            # check if duplicate blocks exist and merge data
            if block_obj.key is None:
                # main @ECLASS block
                if duplicates[tag]:
                    logger.warning(
                        f"'@ECLASS:', line {block_start}: duplicate block")
                duplicates[tag] = True
                # verify name is correct
                file_name = os.path.basename(path)
                if block_data.name != file_name:
                    logger.warning(
                        f"'@ECLASS:' invalid name {block_data.name!r} (should be {file_name!r})")
                data.update(block_data)
            else:
                # item block
                name = block_data['name']
                if name in duplicates[tag]:
                    logger.warning(
                        f'{repr(block[0])}, line {block_start}: duplicate block')
                duplicates[tag].add(name)
                data[block_obj.key].add(block_data)

        return data

    def to_rst(self):
        """Convert eclassdoc object to reStructuredText."""
        if self.name is None:
            raise ValueError('eclass lacking doc support')

        _header_only = partial(_rst_header, newline=True)

        rst = _header_only('=', self.name, leading=True)
        if self.blurb:
            rst.extend(_header_only('-', self.blurb, leading=True))

        if self.description:
            rst.extend(_rst_header('-', 'Description'))
            rst.append(self.description)
            rst.append('')
        if self.deprecated:
            rst.extend(_rst_header('-', 'Deprecated'))
            if isinstance(self.deprecated, bool):
                replacement = 'none'
            else:
                replacement = self.deprecated
            rst.append(f'Replacement: {replacement}')
            rst.append('')
        if self.supported_eapis:
            rst.extend(_rst_header('-', 'Supported EAPIs'))
            rst.append(' '.join(self.supported_eapis))
            rst.append('')
        if self.raw_provides:
            rst.extend(_rst_header('-', 'Transitively Provided Eclasses'))
            rst.append(' '.join(self.raw_provides))
            rst.append('')
        if self.example:
            rst.extend(_rst_header('-', 'Example'))
            rst.append(self.example)
            rst.append('')

        if external_funcs := [x for x in self.functions if not x.internal]:
            rst.extend(_header_only('-', 'Functions'))
            for func in external_funcs:
                header = [func.name]
                if func.usage:
                    header.append(func.usage)
                rst.extend(_rst_header('~', ' '.join(header)))
                if func.description:
                    rst.append(func.description)
                if func.returns:
                    if func.description:
                        rst.append('')
                    rst.append(f'Return value: {func.returns}')
                rst.append('')
        if external_vars := [x for x in self.variables if not x.internal]:
            rst.extend(_header_only('-', 'Variables'))
            for var in external_vars:
                vartype = ''
                if var.required:
                    vartype += ' (REQUIRED)'
                if var.pre_inherit:
                    vartype += ' (SET BEFORE INHERIT)'
                if var.user_variable:
                    vartype += ' (USER VARIABLE)'
                if var.output_variable:
                    vartype += ' (OUTPUT VARIABLE)'

                rst.extend(_rst_header('~', var.name + vartype))
                if var.description:
                    rst.append(var.description)
                rst.append('')
        if external_func_vars := [x for x in self.function_variables if not x.internal]:
            rst.extend(_header_only('-', 'Function Variables'))
            for var in external_func_vars:
                vartype = ''
                if var.required:
                    vartype += ' (REQUIRED)'

                rst.extend(_rst_header('~', var.name + vartype))
                if var.description:
                    rst.append(var.description)
                rst.append('')

        if self.authors:
            rst.extend(_rst_header('-', 'Authors'))
            rst.append('\n'.join(f'| {x}' for x in self.authors))
            rst.append('')
        if self.maintainers:
            rst.extend(_rst_header('-', 'Maintainers'))
            rst.append('\n'.join(f'| {x}' for x in self.maintainers))
            rst.append('')
        if self.bugreports:
            rst.extend(_rst_header('-', 'Bug Reports'))
            rst.append(self.bugreports)
            rst.append('')

        return '\n'.join(rst)

    def _to_docutils(self, writer):
        """Convert eclassdoc object using docutils."""
        from docutils.core import publish_string
        return publish_string(
            source=self.to_rst(), writer=writer,
            settings_overrides={
                'input_encoding': 'unicode',
                'output_encoding': 'unicode',
            }
        )

    def to_man(self):
        """Convert eclassdoc object to a man page."""
        from docutils.writers import manpage

        man_data = {
            'manual_section': '5',
            'manual_group': 'eclass-manpages',
            'date': datetime.utcnow().strftime('%Y-%m-%d'),
            'version': 'Gentoo Linux',
        }
        if self.blurb:
            man_data['subtitle'] = self.blurb

        # add pkgcore version to header comment
        pkgcore_version = get_version(__title__, __file__).split(' --')[0]
        header_comment = f'\nCreated by {pkgcore_version}.'

        class Translator(manpage.Translator):
            """Override docutils man page metadata defaults."""

            document_start = manpage.Translator.document_start + header_comment

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self._docinfo.update(man_data)

        writer = manpage.Writer()
        writer.translator_class = Translator
        return self._to_docutils(writer)

    def to_html(self):
        """Convert eclassdoc object to an HTML 5 document."""
        from docutils.writers import html5_polyglot
        return self._to_docutils(html5_polyglot.Writer())



"""
resolver configuration to match portage behaviour (misbehaviour in a few spots)
"""

__all__ = ["upgrade_resolver", "min_install_resolver"]

from functools import partial
from itertools import chain

from ..repository import misc, multiplex
from ..resolver import plan
from ..restrictions import packages, values
from .atom import atom


def upgrade_resolver(vdbs, dbs, verify_vdb=True, nodeps=False,
                     force_replace=False, resolver_cls=plan.merge_plan,
                     **kwds):

    """
    generate and configure a resolver for upgrading all processed nodes.

    :param vdbs: list of :obj:`pkgcore.repository.prototype.tree` instances
        that represents the livefs
    :param dbs: list of :obj:`pkgcore.repository.prototype.tree` instances
        representing sources of pkgs
    :param verify_vdb: should we stop resolving once we hit the vdb,
        or do full resolution?
    :return: :obj:`pkgcore.resolver.plan.merge_plan` instance
    """

    f = plan.merge_plan.prefer_highest_version_strategy
    # hack.
    if nodeps:
        vdbs = list(map(misc.nodeps_repo, vdbs))
        dbs = list(map(misc.nodeps_repo, dbs))
    elif not verify_vdb:
        vdbs = list(map(misc.nodeps_repo, vdbs))
        dbs = list(dbs)

    if force_replace:
        resolver_cls = generate_replace_resolver_kls(resolver_cls)
    return resolver_cls(dbs + vdbs, plan.pkg_sort_highest, f, **kwds)


def downgrade_resolver(
        vdbs, dbs, verify_vdb=True, nodeps=False, force_replace=False,
        resolver_cls=plan.merge_plan, **kwds):
    """
    generate and configure a resolver for downgrading all processed nodes.

    :param vdbs: list of :obj:`pkgcore.repository.prototype.tree` instances
        that represents the livefs
    :param dbs: list of :obj:`pkgcore.repository.prototype.tree` instances
        representing sources of pkgs
    :param verify_vdb: should we stop resolving once we hit the vdb,
        or do full resolution?
    :return: :obj:`pkgcore.resolver.plan.merge_plan` instance
    """
    restrict = packages.OrRestriction(
        *list(atom(f'>={x.cpvstr}') for x in chain.from_iterable(vdbs)))
    f = partial(plan.merge_plan.prefer_downgrade_version_strategy, restrict)
    dbs = list(map(partial(misc.restrict_repo, restrict), dbs))
    # hack.
    if nodeps:
        vdbs = list(map(misc.nodeps_repo, vdbs))
        dbs = list(map(misc.nodeps_repo, dbs))
    elif not verify_vdb:
        vdbs = list(map(misc.nodeps_repo, vdbs))
        dbs = list(dbs)

    if force_replace:
        resolver_cls = generate_replace_resolver_kls(resolver_cls)
    return resolver_cls(dbs + vdbs, plan.pkg_sort_highest, f, **kwds)


def min_install_resolver(vdbs, dbs, verify_vdb=True, nodeps=False,
                         force_replace=False, resolver_cls=plan.merge_plan,
                         **kwds):
    """
    Resolver that tries to minimize the number of changes while installing.

    generate and configure a resolver that is focused on just
    installing requests- installs highest version it can build a
    solution for, but tries to avoid building anything not needed

    :param vdbs: list of :obj:`pkgcore.repository.prototype.tree` instances
        that represents the livefs
    :param dbs: list of :obj:`pkgcore.repository.prototype.tree` instances
        representing sources of pkgs
    :param verify_vdb: should we stop resolving once we hit the vdb,
        or do full resolution?
    :return: :obj:`pkgcore.resolver.plan.merge_plan` instance
    """

    if nodeps:
        vdbs = list(map(misc.nodeps_repo, vdbs))
        dbs = list(map(misc.nodeps_repo, dbs))
    elif not verify_vdb:
        vdbs = list(map(misc.nodeps_repo, vdbs))
        dbs = list(dbs)

    if force_replace:
        resolver_cls = generate_replace_resolver_kls(resolver_cls)
    return resolver_cls(vdbs + dbs, plan.pkg_sort_highest,
                        plan.merge_plan.prefer_reuse_strategy, **kwds)

_vdb_restrict = packages.OrRestriction(
    packages.PackageRestriction("repo.livefs", values.EqualityMatch(False)),
    packages.AndRestriction(
        packages.PackageRestriction(
            "category", values.StrExactMatch("virtual")),
        packages.PackageRestriction(
            "package_is_real", values.EqualityMatch(False)),
        ),
    )


class empty_tree_merge_plan(plan.merge_plan):

    _vdb_restriction = _vdb_restrict

    def __init__(self, dbs, *args, **kwds):
        """
        :param args: see :obj:`pkgcore.resolver.plan.merge_plan.__init__`
            for valid args
        :param kwds: see :obj:`pkgcore.resolver.plan.merge_plan.__init__`
            for valid args
        """
        super().__init__(dbs, *args, **kwds)
        # XXX *cough*, hack.
        self.default_dbs = multiplex.tree(
            *[x for x in self.all_raw_dbs if not x.livefs])


def generate_replace_resolver_kls(resolver_kls):

    class replace_resolver(resolver_kls):
        overriding_resolver_kls = resolver_kls
        _vdb_restriction = _vdb_restrict

        def add_atoms(self, restricts, **kwds):
            restricts = [packages.KeyedAndRestriction(self._vdb_restriction, x, key=x.key)
                         for x in restricts]
            return self.overriding_resolver_kls.add_atoms(self, restricts, **kwds)

    return replace_resolver



"""
misc. stuff we've not found a spot for yet.
"""

__all__ = (
    "ChunkedDataDict", "IncrementalsDict", "PayloadDict",
    "chunked_data", "collapsed_restrict_to_data", "get_relative_dosym_target",
    "incremental_chunked", "incremental_expansion",
    "incremental_expansion_license",
    "non_incremental_collapsed_restrict_to_data", "optimize_incrementals",
    "sort_keywords",
)

import os.path
from collections import defaultdict, namedtuple
from functools import partial
from itertools import chain

from snakeoil import mappings
from snakeoil.klass import alias_method, generic_equality
from snakeoil.sequences import iflatten_instance

from ..restrictions import boolean, packages, restriction
from . import atom

restrict_payload = namedtuple("restrict_data", ["restrict", "data"])
chunked_data = namedtuple("chunked_data", ("key", "neg", "pos"))


def sort_keywords(keywords):
    """Sort keywords in the proper order: i.e. glob-arches, arch, prefix-arches."""
    def _sort_kwds(kw):
        parts = tuple(reversed(kw.lstrip('~-').partition('-')))
        return parts[0], parts[2]
    return sorted(keywords, key=_sort_kwds)


def optimize_incrementals(sequence):
    # roughly the algorithm walks sequences right->left,
    # identifying terminal points for incrementals; aka, -x x, 'x'
    # is the terminal point- no point in having -x.
    finalized = set()
    for item in reversed(sequence):
        if item[0] == '-':
            i = item[1:]
            if not i:
                raise ValueError("encountered an incomplete negation (just -, no flag)")
            if i == '*':
                # seen enough.
                yield item
                return
            if i not in finalized:
                finalized.add(i)
                yield item
        else:
            if item not in finalized:
                yield item
                finalized.add(item)


def incremental_chunked(orig, iterables):
    for cinst in iterables:
        orig.difference_update(cinst.neg)
        orig.update(cinst.pos)


def incremental_expansion(iterable, orig=None, msg_prefix='', finalize=True):
    if orig is None:
        orig = set()

    for token in iterable:
        if token[0] == '-':
            i = token[1:]
            if not i:
                raise ValueError(
                    f"{msg_prefix} encountered an incomplete negation, '-'")
            if i == '*':
                orig.clear()
            else:
                orig.discard(i)
            if not finalize:
                orig.add(token)
        else:
            orig.discard("-" + token)
            orig.add(token)

    return orig


def incremental_expansion_license(pkg, licenses, license_groups, iterable, msg_prefix=''):
    seen = set()
    for token in iterable:
        if token[0] == '-':
            i = token[1:]
            if not i:
                raise ValueError(
                    f"{pkg}: {msg_prefix}encountered an incomplete negation, '-'")
            if i == '*':
                seen.clear()
            else:
                if i[0] == '@':
                    i = i[1:]
                    if not i:
                        raise ValueError(
                            f"{pkg}: {msg_prefix}encountered an incomplete negation"
                            " of a license group, '-@'")
                    seen.difference_update(license_groups.get(i, ()))
                else:
                    seen.discard(i)
        elif token[0] == '@':
            i = token[1:]
            if not i:
                raise ValueError(
                    f"{pkg}: {msg_prefix}encountered an incomplete license group, '@'")
            seen.update(license_groups.get(i, ()))
        elif token == '*':
            seen.update(licenses)
        else:
            seen.add(token)
    return seen


class IncrementalsDict(mappings.DictMixin):

    disable_py3k_rewriting = True

    def __init__(self, incrementals, **kwds):
        self._incrementals = incrementals
        self._dict = {}
        super().__init__(**kwds)

    def __setitem__(self, key, value):
        if key in self._incrementals:
            if key in self._dict:
                self._dict[key] += f' {value}'
            else:
                self._dict[key] = value
        else:
            self._dict[key] = value

    for x in "getitem delitem len iter".split():
        x = f'__{x}__'
        locals()[x] = alias_method(f"_dict.{x}")
    s = "pop clear keys items values"
    for x in s.split():
        locals()[x] = alias_method(f"_dict.{x}")
    del x, s


class collapsed_restrict_to_data(metaclass=generic_equality):

    __attr_comparison__ = ('defaults', 'freeform', 'atoms', '__class__')
    incremental = True

    def __init__(self, *restrict_sources, **kwds):
        """
        descriptive, no?

        Basically splits an iterable of restrict:data into
        level of specificity, repo, cat, pkg, atom (dict) for use
        in filters

        Finally, a finalize_defaults kwd is supported to control whether
        incremental_expansion finalizes the initial defaults list.
        defaults to True.
        """

        always = []
        repo = []
        cat = []
        pkg = []
        multi = []
        atom_d = {}
        for restrict_pairs in restrict_sources:
            for a, data in restrict_pairs:
                if not data:
                    continue
                if isinstance(a, restriction.AlwaysBool):
                    # yes, odd attr name, but negate holds the val to return.
                    # note also, we're dropping AlwaysFalse; it'll never match.
                    if a.negate:
                        always.extend(data)
                        for atomlist in atom_d.values():
                            atomlist.append(
                                (a, set([flag for flag in data if flag.startswith("-")])))
                elif isinstance(a, atom.atom):
                    atom_d.setdefault(a.key, []).append((a, data))
                elif isinstance(a, boolean.AndRestriction):
                    multi.append((a, data))
                elif isinstance(a, packages.PackageRestriction):
                    if a.attr == "category":
                        cat.append((a, data))
                    elif a.attr == "package":
                        pkg.append((a, data))
                    elif a.attr == "repo.repo_id":
                        repo.append((a, data))
                    else:
                        raise ValueError(
                            f"{a!r} doesn't operate on "
                            f"package/category/repo: data {data!r}")
                else:
                    raise ValueError(
                        f"{a!r} is not an AlwaysBool, PackageRestriction, "
                        f"or atom: data {data!r}")

        if always:
            always = incremental_expansion(always, finalize=kwds.get("finalize_defaults", True))
        else:
            always = set()
        self.defaults = always
        self.defaults_finalized = set(x for x in self.defaults if not x.startswith("-"))
        self.freeform = tuple(x for x in (repo, cat, pkg, multi) if x)
        self.atoms = atom_d

    def pull_data(self, pkg, force_copy=False, pre_defaults=()):
        l = []
        for specific in self.freeform:
            for restrict, data in specific:
                if restrict.match(pkg):
                    l.append(data)
        for atom, data in self.atoms.get(pkg.key, ()):
            if atom.match(pkg):
                l.append(data)

        if pre_defaults:
            s = set(pre_defaults)
            incremental_expansion(self.defaults, orig=s)
        else:
            s = set(self.defaults_finalized)

        if l:
            incremental_expansion(iflatten_instance(l), orig=s)
        return s

    def iter_pull_data(self, pkg, pre_defaults=()):
        for item in pre_defaults:
            yield item
        for item in self.defaults:
            yield item
        for specific in self.freeform:
            for restrict, data in specific:
                if restrict.match(pkg):
                    for item in data:
                        yield item
        for atom, data in self.atoms.get(pkg.key, ()):
            if atom.match(pkg):
                for item in data:
                    yield item


class non_incremental_collapsed_restrict_to_data(collapsed_restrict_to_data):

    def pull_data(self, pkg, force_copy=False):
        l = []
        for specific in self.freeform:
            for restrict, data in specific:
                if restrict.match(pkg):
                    l.append(data)
        for atom, data in self.atoms.get(pkg.key, ()):
            if atom.match(pkg):
                l.append(data)
        if not l:
            if force_copy:
                return set(self.defaults)
            return self.defaults
        s = set(self.defaults)
        s.update(iflatten_instance(l))
        return s

    def iter_pull_data(self, pkg):
        l = [self.defaults]
        for specific in self.freeform:
            l.extend(data for restrict, data in specific if restrict.match(pkg))
        for atom, data in self.atoms.get(pkg.key, ()):
            if atom.match(pkg):
                l.append(data)
        if len(l) == 1:
            return iter(self.defaults)
        return iflatten_instance(l)


def _cached_build_cp_atom_payload(cache, sequence, restrict, payload_form=False):
    sequence = list(sequence)
    key = (payload_form, restrict, tuple(sequence))
    val = cache.get(key)
    if val is None:
        val = cache[key] = _build_cp_atom_payload(sequence, restrict, payload_form=payload_form)
    return val


def _build_cp_atom_payload(sequence, restrict, payload_form=False):
    locked = {}
    ldefault = locked.setdefault

    l = []

    if payload_form:
        def f(r, neg, pos):
            return restrict_payload(r, tuple(chain(('-' + x for x in neg), pos)))
    else:
        f = chunked_data

    i = list(sequence)
    if len(i) <= 1:
        if not i:
            return ()
        return (f(i[0].key, i[0].neg, i[0].pos),)

    i = reversed(i)

    for data in i:
        if data.key == packages.AlwaysTrue or getattr(data.key, 'is_simple', False):
            for n in data.neg:
                ldefault(n, False)
            for p in data.pos:
                ldefault(p, True)
            continue
        neg = tuple(x for x in data.neg if x not in locked)
        pos = tuple(x for x in data.pos if x not in locked)
        if neg or pos:
            l.append((data.key, neg, pos))

    # thus far we've done essentially a tracing for R->L, of globals,
    # this leaves d-u/a X, =d-u/a-1 X # slipping through however,
    # since the specific is later.  Plus it's reversed from what we want.
    # so we rebuild, but apply the same global trick as we go.

    if not locked:
        # all is specific/non-simple, just reverse and return
        return tuple(f(*vals) for vals in reversed(l))

    new_l = [f(
        restrict,
        tuple(k for k, v in locked.items() if not v),  # neg
        tuple(k for k, v in locked.items() if v)  # pos
    )]
    # we exploit a few things this time around in reusing the algo from above
    # we know there is only going to be one global (which we just added),
    # and that everything is specific.

    lget = locked.get

    for key, neg, pos in reversed(l):
        # only grab the deltas; if a + becomes a specific -
        neg = tuple(x for x in neg if lget(x, True))
        pos = tuple(x for x in pos if not lget(x, False))
        if neg or pos:
            new_l.append(f(key, neg, pos))

    return tuple(new_l)


class ChunkedDataDict(metaclass=generic_equality):

    __attr_comparison__ = ('_global_settings', '_dict')

    def __init__(self):
        self._global_settings = []
        self._dict = defaultdict(partial(list, self._global_settings))

    @property
    def frozen(self):
        return isinstance(self._dict, mappings.ImmutableDict)

    def clone(self, unfreeze=False):
        obj = self.__class__()
        if self.frozen and not unfreeze:
            obj._dict = self._dict
            obj._global_settings = self._global_settings
            return obj
        obj._dict = defaultdict(partial(list, self._global_settings))
        for key, values in self._dict.items():
            obj._dict[key].extend(values)
        obj._global_settings = list(self._global_settings)
        return obj

    def mk_item(self, key, neg, pos):
        return chunked_data(key, tuple(neg), tuple(pos))

    def add_global(self, item):
        return self._add_global(item.neg, item.pos, restrict=item.key)

    def add_bare_global(self, disabled, enabled):
        return self._add_global(disabled, enabled)

    def _add_global(self, disabled, enabled, restrict=None):
        if not disabled and not enabled:
            return
        # discard current global in the mapping.
        disabled = set(disabled)
        enabled = set(enabled)
        if restrict is None:
            restrict = packages.AlwaysTrue
        payload = self.mk_item(restrict, tuple(disabled), tuple(enabled))
        for vals in self._dict.values():
            vals.append(payload)

        self._expand_globals([payload])

    def merge(self, cdict):
        if not isinstance(cdict, ChunkedDataDict):
            raise TypeError(
                "merge expects a ChunkedDataDict instance; "
                f"got type {type(cdict)}, {cdict!r}")
        if isinstance(cdict, PayloadDict) and not isinstance(self, PayloadDict):
            raise TypeError(
                "merge expects a PayloadDataDict instance; "
                f"got type {type(cdict)}, {cdict!r}")
        # straight extensions for this, rather than update_from_stream.
        d = self._dict
        for key, values in cdict._dict.items():
            d[key].extend(values)

        # note the cdict we're merging has the globals layer through it already, ours
        # however needs to have the new globals appended to all untouched keys
        # (no need to update the merged keys- they already have that global data
        # interlaced)
        new_globals = cdict._global_settings
        if new_globals:
            updates = set(d)
            updates.difference_update(cdict._dict)
            for key in updates:
                d[key].extend(new_globals)
            self._expand_globals(new_globals)

    def _expand_globals(self, new_globals):
        # while a chain seems obvious here, reversed is used w/in _build_cp_atom;
        # reversed doesn't like chain, so we just modify the list and do it this way.
        self._global_settings.extend(new_globals)
        restrict = getattr(new_globals[0], 'key', packages.AlwaysTrue)
        if restrict == packages.AlwaysTrue:
            self._global_settings[:] = list(
                _build_cp_atom_payload(self._global_settings, restrict))

    def add(self, cinst):
        self.update_from_stream([cinst])

    def update_from_stream(self, stream):
        for cinst in stream:
            if getattr(cinst.key, 'key', None) is not None:
                # atom, or something similar.  use the key lookup.
                # hack also... recreate the restriction; this is due to
                # internal idiocy in ChunkedDataDict that will be fixed.
                new_globals = (x for x in self._global_settings
                               if x not in self._dict[cinst.key.key])
                self._dict[cinst.key.key].extend(new_globals)
                self._dict[cinst.key.key].append(cinst)
            else:
                self.add_global(cinst)

    def freeze(self):
        if not isinstance(self._dict, mappings.ImmutableDict):
            self._dict = mappings.ImmutableDict(
                (k, tuple(v))
                for k, v in self._dict.items())
            self._global_settings = tuple(self._global_settings)

    def optimize(self, cache=None):
        if cache is None:
            d_stream = (
                (k, _build_cp_atom_payload(v, atom.atom(k), False))
                for k, v in self._dict.items())
            g_stream = (_build_cp_atom_payload(
                self._global_settings,
                packages.AlwaysTrue, payload_form=isinstance(self, PayloadDict)))
        else:
            d_stream = ((k, _cached_build_cp_atom_payload(
                cache, v, atom.atom(k), False))
                for k, v in self._dict.items())
            g_stream = (_cached_build_cp_atom_payload(
                cache, self._global_settings,
                packages.AlwaysTrue, payload_form=isinstance(self, PayloadDict)))

        if self.frozen:
            self._dict = mappings.ImmutableDict(d_stream)
            self._global_settings = tuple(g_stream)
        else:
            self._dict.update(d_stream)
            self._global_settings[:] = list(g_stream)

    def render_to_dict(self):
        d = dict(self._dict)
        if self._global_settings:
            d[packages.AlwaysTrue] = self._global_settings[:]
        return d

    def render_to_payload(self):
        d = PayloadDict()
        d = {atom.atom(k): _build_cp_atom_payload(v, atom.atom(k), True)
             for k, v in self._dict.items()}
        if self._global_settings:
            data = _build_cp_atom_payload(
                self._global_settings,
                packages.AlwaysTrue, payload_form=True)
            d[packages.AlwaysTrue] = tuple(data)
        return d

    def __bool__(self):
        return bool(self._global_settings) or bool(self._dict)

    def __str__(self):
        return str(self.render_to_dict())

    def render_pkg(self, pkg, pre_defaults=()):
        items = self._dict.get(pkg.key)
        if items is None:
            items = self._global_settings
        s = set(pre_defaults)
        incremental_chunked(s, (cinst for cinst in items if cinst.key.match(pkg)))
        return s

    pull_data = render_pkg


class PayloadDict(ChunkedDataDict):

    def mk_item(self, key, neg, pos):
        return restrict_payload(key, tuple(chain(("-" + x for x in neg), pos)))

    def add_bare_global(self, payload):
        neg = [x[1:] for x in payload if x[0] == '-']
        pos = [x for x in payload if x[0] != '-']
        ChunkedDataDict.add_bare_global(self, neg, pos)

    def add_global(self, pinst):
        neg = [x[1:] for x in pinst.data if x[0] == '-']
        pos = [x for x in pinst.data if x[0] != '-']
        return ChunkedDataDict.add_global(self, chunked_data(pinst.restrict, neg, pos))

    def update_from_stream(self, stream):
        for pinst in stream:
            if getattr(pinst.restrict, 'key', None) is not None:
                # atom, or something similar.  use the key lookup.
                # hack also... recreate the restriction; this is due to
                # internal idiocy in ChunkedDataDict that will be fixed.
                self._dict[pinst.restrict.key].append(pinst)
            else:
                self.add_global(pinst)

    def render_pkg(self, pkg, pre_defaults=()):
        items = self._dict.get(atom.atom(pkg.key))
        if items is None:
            items = self._global_settings
        s = set(pre_defaults)
        data = chain.from_iterable(
            item.data for item in items if item.restrict.match(pkg))
        return incremental_expansion(data, orig=s)

    pull_data = render_pkg


def run_sanity_checks(pkgs, domain, threads=None):
    """Run all sanity checks for a sequence of packages."""
    sanity_failures = defaultdict(list)
    # TODO: parallelize this across separate processes
    for pkg in pkgs:
        pkg_ops = domain.pkg_operations(pkg)
        if pkg_ops.supports("sanity_check") and (failures := pkg_ops.sanity_check()):
            sanity_failures[pkg] = failures
    return sanity_failures


def get_relative_dosym_target(source, target):
    """Get relative path from target to source, for symlink target."""
    # NB: as dosym arg, initial slash can be omitted
    return os.path.relpath(source, os.path.join('/', os.path.dirname(target)))



import os

from pkgcore.fs import fs, livefs
from pkgcore.fs.contents import contentsSet
from snakeoil.osutils import pjoin
from snakeoil.test import TestCase
from snakeoil.test.mixins import TempDirMixin


class FsObjsTest(TempDirMixin, TestCase):

    def check_attrs(self, obj, path, offset=None):
        if offset is None:
            st = os.lstat(path)
        else:
            st = os.lstat(offset + '/' + path)
        if offset is not None:
            self.assertTrue(
                path.startswith("/"),
                msg=f"path must be absolute, got {path!r}")
        self.assertEqual(obj.mode & 0o7777, st.st_mode & 0o7777)
        self.assertEqual(obj.uid, st.st_uid)
        self.assertEqual(obj.gid, st.st_gid)
        if fs.isreg(obj):
            if offset is None:
                self.assertEqual(obj.data.path, path)
            else:
                self.assertEqual(obj.data.path,
                    offset + path)

    def test_data_source(self):
        o = livefs.gen_obj("/tmp/etc/passwd", real_location="/etc/passwd")
        self.assertTrue(o.location, "/tmp/etc/passwd")
        self.assertTrue(o.data.path, "/etc/passwd")
        with open("/etc/passwd", "rb") as f:
            self.assertTrue(o.data.bytes_fileobj().read(), f.read())

    def test_gen_obj_reg(self):
        path = os.path.join(self.dir, "reg_obj")
        open(path, "w").close()
        o = livefs.gen_obj(path)
        self.assertTrue(fs.isreg(o))
        self.check_attrs(o, path)
        o2 = livefs.gen_obj(path, inode=None)
        self.check_attrs(o, path)
        self.assertNotEqual(o.inode, o2.inode)

    def test_gen_obj_dir(self):
        o = livefs.gen_obj(self.dir)
        self.assertTrue(fs.isdir(o))
        self.check_attrs(o, self.dir)

    def test_gen_obj_sym(self):
        path = os.path.join(self.dir, "test_sym")
        os.mkdir(path)
        src = os.path.join(path, "s")
        link = os.path.join(path, "t")
        open(src, "w").close()
        os.symlink(src, link)
        obj = livefs.gen_obj(link)
        self.assertInstance(obj, fs.fsSymlink)
        self.check_attrs(obj, link)
        self.assertEqual(os.readlink(link), obj.target)

    def test_gen_obj_fifo(self):
        path = os.path.join(self.dir, "fifo")
        os.mkfifo(path)
        o = livefs.gen_obj(path)
        self.check_attrs(o, path)

    def test_iterscan(self):
        path = os.path.join(self.dir, "iscan")
        os.mkdir(path)
        files = [os.path.normpath(os.path.join(path, x)) for x in [
                "tmp", "blah", "dar"]]
        # cheap version of a touch.
        for x in files:
            open(x, "w").close()
        dirs = [os.path.normpath(os.path.join(path, x)) for x in [
                "a", "b", "c"]]
        for x in dirs:
            os.mkdir(x)
        dirs.append(path)
        for obj in livefs.iter_scan(path):
            self.assertInstance(obj, fs.fsBase)
            if fs.isreg(obj):
                self.assertTrue(obj.location in files)
            elif fs.isdir(obj):
                self.assertTrue(obj.location in dirs)
            else:
                raise Exception(f"unknown object popped up in testing dir, {obj!r}")
            self.check_attrs(obj, obj.location)
        # do offset verification now.
        offset = os.path.join(self.dir, "iscan")
        for obj in livefs.iter_scan(path, offset=offset):
            self.check_attrs(obj, obj.location, offset=offset)

        seen = []
        for obj in livefs.iter_scan(files[0]):
            self.check_attrs(obj, obj.location)
            seen.append(obj.location)
        self.assertEqual((files[0],), tuple(sorted(seen)))

    def test_sorted_scan(self):
        path = os.path.join(self.dir, "sorted_scan")
        os.mkdir(path)
        files = [os.path.normpath(os.path.join(path, x)) for x in
                 ["tmp", "blah", "dar"]]
        # cheap version of a touch.
        for x in files:
            open(x, "w").close()
        dirs = [os.path.normpath(os.path.join(path, x)) for x in [
                "a", "b", "c"]]
        for x in dirs:
            os.mkdir(x)

        # regular directory scanning
        sorted_files = livefs.sorted_scan(path)
        self.assertEqual(
            list([pjoin(path, x) for x in ['blah', 'dar', 'tmp']]),
            sorted_files)

        # nonexistent paths
        nonexistent_path = os.path.join(self.dir, 'foobar')
        sorted_files = livefs.sorted_scan(nonexistent_path)
        self.assertEqual(sorted_files, [])
        sorted_files = livefs.sorted_scan(nonexistent_path, nonexistent=True)
        self.assertEqual(sorted_files, [nonexistent_path])

    def test_sorted_scan_hidden(self):
        path = os.path.join(self.dir, "sorted_scan")
        os.mkdir(path)
        files = [os.path.normpath(os.path.join(path, x)) for x in
                 [".tmp", "blah",]]
        # cheap version of a touch.
        for x in files:
            open(x, "w").close()
        sorted_files = livefs.sorted_scan(path)
        assert list([pjoin(path, x) for x in ['.tmp', 'blah']]) == sorted_files
        sorted_files = livefs.sorted_scan(path, hidden=False)
        assert list([pjoin(path, x) for x in ['blah']]) == sorted_files

    def test_sorted_scan_backup(self):
        path = os.path.join(self.dir, "sorted_scan")
        os.mkdir(path)
        files = [os.path.normpath(os.path.join(path, x)) for x in ["blah", "blah~"]]
        # cheap version of a touch.
        for x in files:
            open(x, "w").close()
        sorted_files = livefs.sorted_scan(path)
        assert list([pjoin(path, x) for x in ['blah', 'blah~']]) == sorted_files
        sorted_files = livefs.sorted_scan(path, backup=False)
        assert list([pjoin(path, x) for x in ['blah']]) == sorted_files

    def test_relative_sym(self):
        f = os.path.join(self.dir, "relative-symlink-test")
        os.symlink("../sym1/blah", f)
        o = livefs.gen_obj(f)
        self.assertTrue(o.target == "../sym1/blah")

    def test_intersect(self):
        open(pjoin(self.dir, 'reg'), 'w').close()
        cset = contentsSet([fs.fsFile('reg', strict=False)])
        cset = cset.insert_offset(self.dir)
        self.assertEqual(contentsSet(livefs.intersect(cset)), cset)
        cset = contentsSet([fs.fsFile('reg/foon', strict=False),
            fs.fsFile('reg/dar', strict=False),
            fs.fsDir('reg/dir', strict=False)]).insert_offset(self.dir)
        self.assertEqual(list(livefs.intersect(cset)), [])
        cset = contentsSet([fs.fsDir('reg', strict=False)])
        self.assertEqual(list(livefs.intersect(cset)), [])



"""system/repository maintenance utility"""

import argparse
import logging
import os
import textwrap
import time
from multiprocessing import cpu_count

from snakeoil.cli import arghparse
from snakeoil.contexts import patch
from snakeoil.fileutils import AtomicWriteFile
from snakeoil.osutils import pjoin
from snakeoil.sequences import iter_stable_unique

from ..cache.flat_hash import md5_cache
from ..ebuild import repository as ebuild_repo
from ..ebuild import triggers
from ..ebuild.cpv import CPV
from ..ebuild.eclass import EclassDoc
from ..exceptions import PkgcoreUserException
from ..fs import contents, livefs
from ..merge import triggers as merge_triggers
from ..operations import OperationError
from ..operations import observer as observer_mod
from ..package import mutated
from ..package.errors import MetadataException
from ..util import commandline

pkgcore_opts = commandline.ArgumentParser(domain=False, script=(__file__, __name__))
argparser = commandline.ArgumentParser(
    suppress=True, description=__doc__, parents=(pkgcore_opts,))
subparsers = argparser.add_subparsers(description="general system maintenance")

shared_options = (commandline.ArgumentParser(
    config=False, color=False, debug=False, quiet=False, verbose=False,
    version=False, domain=False, add_help=False),)
shared_options_domain = (commandline.ArgumentParser(
    config=False, color=False, debug=False, quiet=False, verbose=False,
    version=False, domain=True, add_help=False),)

sync = subparsers.add_parser(
    "sync", parents=shared_options,
    description="synchronize a local repository with its defined remote")
sync.add_argument(
    'repos', metavar='repo', nargs='*', help="repo(s) to sync",
    action=commandline.StoreRepoObject, store_name=True, repo_type='config')
sync.add_argument(
    '-f', '--force', action='store_true', default=False,
    help="force syncing to occur regardless of staleness checks")
@sync.bind_main_func
def sync_main(options, out, err):
    """Update local repos to match their remotes."""
    succeeded, failed = [], []

    for repo_name, repo in iter_stable_unique(options.repos):
        # rewrite the name if it has the usual prefix
        if repo_name.startswith("conf:"):
            repo_name = repo_name[5:]

        if not repo.operations.supports("sync"):
            continue
        out.write(f"*** syncing {repo_name}")
        ret = False
        err_msg = ''
        # repo operations don't yet take an observer, thus flush
        # output to keep lines consistent.
        out.flush()
        err.flush()
        try:
            ret = repo.operations.sync(
                force=options.force, verbosity=options.verbosity)
        except OperationError as e:
            exc = getattr(e, '__cause__', e)
            if not isinstance(exc, PkgcoreUserException):
                raise
            err_msg = f': {exc}'
        if not ret:
            out.write(f"!!! failed syncing {repo_name}{err_msg}")
            failed.append(repo_name)
        else:
            succeeded.append(repo_name)
            out.write(f"*** synced {repo_name}")

    out.flush()
    err.flush()
    total = len(succeeded) + len(failed)
    if total > 1:
        results = []
        succeeded = ', '.join(sorted(succeeded))
        failed = ', '.join(sorted(failed))
        if succeeded:
            results.append(f"*** synced: {succeeded}")
        if failed:
            results.append(f"!!! failed: {failed}")
        results = "\n".join(results)
        out.write(f"\n*** sync results:\n{results}")
    return 1 if failed else 0


# TODO: restrict to required repo types
copy = subparsers.add_parser(
    "copy", parents=shared_options_domain,
    description="copy binpkgs between repos; primarily useful for "
    "quickpkging a livefs pkg")
copy.add_argument(
    'target_repo', action=commandline.StoreRepoObject, repo_type='binary-raw',
    writable=True, help="repository to add packages to")
commandline.make_query(
    copy, nargs='+', dest='query',
    help="packages matching any of these restrictions will be selected "
    "for copying")
copy_opts = copy.add_argument_group("subcommand options")
copy_opts.add_argument(
    '-s', '--source-repo', default=None, repo_type='installed',
    action=commandline.StoreRepoObject,
    help="copy strictly from the supplied repository; else it copies from "
    "wherever a match is found")
copy_opts.add_argument(
    '-i', '--ignore-existing', default=False, action='store_true',
    help="if a matching pkg already exists in the target, don't update it")
@copy.bind_main_func
def copy_main(options, out, err):
    """Copy pkgs between repos."""
    source_repo = options.source_repo
    if source_repo is None:
        source_repo = options.domain.all_source_repos
    target_repo = options.target_repo

    failures = False

    for pkg in source_repo.itermatch(options.query):
        if options.ignore_existing and pkg.versioned_atom in target_repo:
            out.write(f"skipping existing pkg: {pkg.cpvstr}")
            continue
        # TODO: remove this once we limit src repos to non-virtual (pkg.provided) repos
        if not getattr(pkg, 'package_is_real', True):
            out.write(f"skipping virtual pkg: {pkg.cpvstr}")
            continue

        out.write(f"copying {pkg}... ")
        if getattr(getattr(pkg, 'repo', None), 'livefs', False):
            out.write("forcing regen of contents due to src being livefs..")
            new_contents = contents.contentsSet(mutable=True)
            for fsobj in pkg.contents:
                try:
                    new_contents.add(livefs.gen_obj(fsobj.location))
                except FileNotFoundError:
                    err.write(
                        f"warning: dropping fs obj {fsobj!r} since it doesn't exist")
                except OSError as oe:
                    err.write(
                        f"failed accessing fs obj {fsobj!r}; {oe}\n"
                        "aborting this copy")
                    failures = True
                    new_contents = None
                    break
            if new_contents is None:
                continue
            pkg = mutated.MutatedPkg(pkg, {'contents': new_contents})

        target_repo.operations.install_or_replace(pkg).finish()
        out.write("completed\n")

    if failures:
        return 1
    return 0


def _get_default_jobs(namespace, attr):
    # we intentionally overschedule for SMP; the main python thread
    # isn't too busy, thus we want to keep all bash workers going.
    val = cpu_count()
    if val > 1:
        val += 1
    setattr(namespace, attr, val)


def update_use_local_desc(repo, observer):
    """Update a repo's local USE flag description cache (profiles/use.local.desc)"""
    ret = 0
    use_local_desc = pjoin(repo.location, "profiles", "use.local.desc")
    f = None

    def _raise_xml_error(exc):
        observer.error(f'{cat}/{pkg}: failed parsing metadata.xml: {str(exc)}')
        nonlocal ret
        ret = 1

    try:
        f = AtomicWriteFile(use_local_desc)
        f.write(textwrap.dedent('''\
            # This file is deprecated as per GLEP 56 in favor of metadata.xml.
            # Please add your descriptions to your package's metadata.xml ONLY.
            # * generated automatically using pmaint *\n\n'''))
        with patch('pkgcore.log.logger.error', _raise_xml_error):
            for cat, pkgs in sorted(repo.packages.items()):
                for pkg in sorted(pkgs):
                    metadata = repo._get_metadata_xml(cat, pkg)
                    for flag, desc in sorted(metadata.local_use.items()):
                        f.write(f'{cat}/{pkg}:{flag} - {desc}\n')
        f.close()
    except IOError as e:
        observer.error(
            f"Unable to update use.local.desc file {use_local_desc!r}: {e.strerror}")
        ret = os.EX_IOERR
    finally:
        if f is not None:
            f.discard()

    return ret


def update_pkg_desc_index(repo, observer):
    """Update a repo's package description cache (metadata/pkg_desc_index)"""
    ret = 0
    pkg_desc_index = pjoin(repo.location, "metadata", "pkg_desc_index")
    f = None
    try:
        f = AtomicWriteFile(pkg_desc_index)
        for cat, pkgs in sorted(repo.packages.items()):
            for pkg in sorted(pkgs):
                cpvs = sorted(CPV(cat, pkg, v) for v in repo.versions[(cat, pkg)])
                # get the most recent pkg description, skipping bad pkgs
                for cpv in reversed(cpvs):
                    try:
                        desc = repo[(cat, pkg, cpv.fullver)].description
                        versions = ' '.join(x.fullver for x in cpvs)
                        f.write(f"{cat}/{pkg} {versions}: {desc}\n")
                        break
                    except MetadataException as e:
                        # should be caught and outputted already by cache regen
                        ret = 1
        f.close()
    except IOError as e:
        observer.error(
            f"Unable to update pkg_desc_index file {pkg_desc_index!r}: {e.strerror}")
        ret = os.EX_IOERR
    finally:
        if f is not None:
            f.discard()

    return ret


regen = subparsers.add_parser(
    "regen", parents=shared_options_domain,
    description="regenerate repository caches")
regen.add_argument(
    'repos', metavar='repo', nargs='*',
    action=commandline.StoreRepoObject, repo_type='source-raw', allow_external_repos=True,
    help="repo(s) to regenerate caches for")
regen_opts = regen.add_argument_group("subcommand options")
regen_opts.add_argument(
    "--disable-eclass-caching", action='store_true', default=False,
    help="""
        For regen operation, pkgcore internally turns on an optimization that
        caches eclasses into individual functions thus parsing the eclass only
        twice max per EBD processor. Disabling this optimization via this
        option results in ~2x slower regeneration. Disable it only if you
        suspect the optimization is somehow causing issues.
    """)
regen_opts.add_argument(
    "-t", "--threads", type=int,
    default=arghparse.DelayedValue(_get_default_jobs, 100),
    help="number of threads to use",
    docs="""
        Number of threads to use for regeneration, defaults to using all
        available processors.
    """)
regen_opts.add_argument(
    "--force", action='store_true', default=False,
    help="force regeneration to occur regardless of staleness checks or repo settings")
regen_opts.add_argument(
    "--dir", dest='cache_dir', type=arghparse.create_dir,
    help="use separate directory to store repository caches")
regen_opts.add_argument(
    "--rsync", action='store_true', default=False,
    help="perform actions necessary for rsync repos (update metadata/timestamp.chk)")
regen_opts.add_argument(
    "--use-local-desc", action='store_true', default=False,
    help="update local USE flag description cache (profiles/use.local.desc)")
regen_opts.add_argument(
    "--pkg-desc-index", action='store_true', default=False,
    help="update package description cache (metadata/pkg_desc_index)")
@regen.bind_main_func
def regen_main(options, out, err):
    """Regenerate a repository cache."""
    ret = []

    observer = observer_mod.formatter_output(out)
    for repo in iter_stable_unique(options.repos):
        if options.cache_dir is not None:
            # recreate new repo object with cache dir override
            cache = (md5_cache(pjoin(options.cache_dir.rstrip(os.sep), repo.repo_id)),)
            repo = ebuild_repo.tree(
                options.config, repo.config, cache=cache)
        if not repo.operations.supports("regen_cache"):
            out.write(f"repo {repo} doesn't support cache regeneration")
            continue
        elif not getattr(repo, 'cache', False) and not options.force:
            out.write(f"skipping repo {repo}: cache disabled")
            continue

        start_time = time.time()
        ret.append(repo.operations.regen_cache(
            threads=options.threads, observer=observer, force=options.force,
            eclass_caching=(not options.disable_eclass_caching)))
        end_time = time.time()

        if options.verbosity > 0:
            out.write(
                "finished %d nodes in %.2f seconds" %
                (len(repo), end_time - start_time))

        if options.rsync:
            timestamp = pjoin(repo.location, "metadata", "timestamp.chk")
            try:
                with open(timestamp, "w") as f:
                    f.write(time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.gmtime()))
            except IOError as e:
                err.write(f"Unable to update timestamp file {timestamp!r}: {e.strerror}")
                ret.append(os.EX_IOERR)

        if options.use_local_desc:
            ret.append(update_use_local_desc(repo, observer))
        if options.pkg_desc_index:
            ret.append(update_pkg_desc_index(repo, observer))

    return int(any(ret))


env_update = subparsers.add_parser(
    "env-update", description="update env.d and ldconfig",
    parents=shared_options_domain)
env_update_opts = env_update.add_argument_group("subcommand options")
env_update_opts.add_argument(
    "--skip-ldconfig", action='store_true', default=False,
    help="do not update etc/ldso.conf and ld.so.cache")
@env_update.bind_main_func
def env_update_main(options, out, err):
    root = getattr(options.domain, 'root', None)
    if root is None:
        env_update.error("domain specified lacks a root setting; is it a virtual or remote domain?")

    out.write(f"updating env for {root!r}...")
    try:
        triggers.perform_env_update(root, skip_ldso_update=options.skip_ldconfig)
    except PermissionError:
        env_update.error("failed updating env, lacking permissions")
    if not options.skip_ldconfig:
        out.write(f"update ldso cache/elf hints for {root!r}...")
        merge_triggers.update_elf_hints(root)
    return 0


class EclassArgs(argparse.Action):
    """Determine eclass arguments for `pmaint eclass`."""

    def __call__(self, parser, namespace, values, option_string=None):
        if values:
            eclasses = []
            for val in values:
                path = os.path.realpath(val)
                if os.path.isdir(path):
                    eclasses.extend(os.listdir(path))
                elif val.endswith('.eclass'):
                    eclasses.append(path)
                else:
                    raise argparse.ArgumentError(self, f'invalid eclass: {val!r}')
            eclasses = sorted(x for x in eclasses if x.endswith('.eclass'))
        else:
            eclass_dir = pjoin(namespace.repo.location, 'eclass')
            try:
                files = sorted(os.listdir(eclass_dir))
            except FileNotFoundError:
                files = []
            eclasses = [pjoin(eclass_dir, x) for x in files if x.endswith('.eclass')]
            if not eclasses:
                parser.error(f'{namespace.repo.repo_id} repo: no eclasses found')

        setattr(namespace, self.dest, eclasses)


eclass = subparsers.add_parser(
    "eclass", parents=shared_options_domain,
    description="generate eclass docs")
eclass.add_argument(
    'eclasses', nargs='*', help="eclasses to target",
    action=arghparse.Delayed, target=EclassArgs, priority=1001)
eclass_opts = eclass.add_argument_group("subcommand options")
eclass_opts.add_argument(
    "--dir", dest='output_dir', type=arghparse.create_dir, help="output directory")
eclass_opts.add_argument(
    "-f", "--format", help="output format",
    default='man', choices=('rst', 'man', 'html'))
eclass_opts.add_argument(
    "-r", "--repo", help="target repository",
    action=commandline.StoreRepoObject, repo_type='ebuild-raw', allow_external_repos=True,
    docs="""
        Target repository to search for eclasses. If no repo is specified the default repo is used.
    """)


@eclass.bind_delayed_default(1000, 'repo')
def _eclass_default_repo(namespace, attr):
    """Use default repo if none is selected."""
    repo = namespace.config.get_default('repo')
    setattr(namespace, attr, repo)


@eclass.bind_delayed_default(1000, 'output_dir')
def _eclass_default_output_dir(namespace, attr):
    """Use CWD as output dir if unset."""
    setattr(namespace, attr, os.getcwd())


@eclass.bind_main_func
def _eclass_main(options, out, err):
    # suppress all eclassdoc parsing warnings
    logging.getLogger('pkgcore').setLevel(100)
    failed = []

    # determine output file extension
    ext_map = {'man': '5'}
    ext = ext_map.get(options.format, options.format)

    for path in options.eclasses:
        try:
            with open(pjoin(options.output_dir, f'{os.path.basename(path)}.{ext}'), 'wt') as f:
                obj = EclassDoc(path)
                convert_func = getattr(obj, f'to_{options.format}')
                f.write(convert_func())
        except ValueError as e:
            # skip eclasses lacking eclassdoc support
            err.write(f'{eclass.prog}: skipping {path!r}: {e}')
            err.flush()
        except IOError as e:
            err.write(f'{eclass.prog}: error: {path!r}: {e}')
            err.flush()
            failed.append(path)

    return int(any(failed))



import os
import subprocess
import textwrap
from collections.abc import MutableSet
from datetime import datetime

from snakeoil import klass
from snakeoil.fileutils import touch
from snakeoil.osutils import pjoin

import pytest


class GitRepo:
    """Class for creating/manipulating git repos.

    Only relies on the git binary existing in order to limit
    dependency requirements.
    """

    def __init__(self, path, bare=False, branch='main', commit=False, clone=False):
        self.path = path
        if clone:
            os.makedirs(self.path)
            self.run(['git', 'clone', clone, self.path])
        else:
            self.run(['git', 'init', '-b', branch] + (['--bare'] if bare else []) + [self.path])
            self.run(['git', 'config', 'user.email', 'first.last@email.com'])
            self.run(['git', 'config', 'user.name', 'First Last'])

        if commit:
            if self.changes:
                # if files exist in the repo, add them in an initial commit
                self.add_all(msg='initial commit')
            else:
                # otherwise add a stub initial commit
                self.add(pjoin(self.path, '.init'), create=True)

    def run(self, cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, **kwargs):
        return subprocess.run(
            cmd, cwd=self.path, encoding='utf8', check=True,
            stdout=stdout, stderr=stderr, **kwargs)

    def log(self, args):
        """Run ``git log`` with given args and return a list of outputted lines."""
        p = self.run(['git', 'log'] + args, stdout=subprocess.PIPE)
        return p.stdout.strip().splitlines()

    @property
    def changes(self):
        """Return a list of any untracked or modified files in the repo."""
        cmd = ['git', 'ls-files', '-mo', '--exclude-standard']
        p = self.run(cmd, stdout=subprocess.PIPE)
        return p.stdout.splitlines()

    @property
    def HEAD(self):
        """Return the commit hash for git HEAD."""
        p = self.run(['git', 'rev-parse', '--short', 'HEAD'], stdout=subprocess.PIPE)
        return p.stdout.strip()

    def __str__(self):
        return self.path

    def commit(self, msg, signoff=False):
        """Make a commit to the repo."""
        if isinstance(msg, str):
            msg = msg.splitlines()
        if signoff:
            msg.extend(['', 'Signed-off-by: First Last <first.last@email.com>'])
        self.run(['git', 'commit', '-m', '\n'.join(msg)])

    def add(self, file_path, msg='commit', commit=True, create=False, signoff=False):
        """Add a file and commit it to the repo."""
        if create:
            touch(pjoin(self.path, file_path))
        self.run(['git', 'add', file_path])
        if commit:
            self.commit(msg, signoff)

    def add_all(self, msg='commit-all', commit=True, signoff=False):
        """Add and commit all tracked and untracked files."""
        self.run(['git', 'add', '--all'])
        if commit:
            self.commit(msg, signoff)

    def remove(self, path, msg='remove', commit=True, signoff=False):
        """Remove a given file path and commit the change."""
        self.run(['git', 'rm', path])
        if commit:
            self.commit(msg, signoff)

    def remove_all(self, path, msg='remove-all', commit=True, signoff=False):
        """Remove all files from a given path and commit the changes."""
        self.run(['git', 'rm', '-rf', path])
        if commit:
            self.commit(msg, signoff)

    def move(self, path, new_path, msg=None, commit=True, signoff=False):
        """Move a given file path and commit the change."""
        msg = msg if msg is not None else f'{path} -> {new_path}'
        self.run(['git', 'mv', path, new_path])
        if commit:
            self.commit(msg, signoff)


@pytest.fixture
def git_repo(tmp_path_factory):
    """Create an empty git repo with an initial commit."""
    return GitRepo(str(tmp_path_factory.mktemp('git-repo')), commit=True)


@pytest.fixture
def make_git_repo(tmp_path_factory):
    """Factory for git repo creation."""
    def _make_git_repo(path=None, **kwargs):
        path = str(tmp_path_factory.mktemp('git-repo')) if path is None else path
        return GitRepo(path, **kwargs)
    return _make_git_repo


class _FileSet(MutableSet):
    """Set object that maps to file content updates for a given path."""

    def __init__(self, path):
        self._path = path
        self._set = set()

    def _sync(self):
        with open(self._path, 'w') as f:
            f.write('\n'.join(self._set) + '\n')

    def __contains__(self, key):
        return key in self._set

    def __iter__(self):
        return iter(self._set)

    def __len__(self):
        return len(self._set)

    def update(self, iterable):
        orig_entries = len(self._set)
        self._set.update(iterable)
        if len(self._set) != orig_entries:
            self._sync()

    def add(self, value):
        orig_entries = len(self._set)
        self._set.add(value)
        if len(self._set) != orig_entries:
            self._sync()

    def remove(self, value):
        orig_entries = len(self._set)
        self._set.remove(value)
        if len(self._set) != orig_entries:
            self._sync()

    def discard(self, value):
        orig_entries = len(self._set)
        self._set.discard(value)
        if len(self._set) != orig_entries:
            self._sync()


class EbuildRepo:
    """Class for creating/manipulating ebuild repos."""

    def __init__(self, path, repo_id='fake', eapi='5', masters=(), arches=()):
        self.path = path
        self.arches = _FileSet(pjoin(self.path, 'profiles', 'arch.list'))
        self._today = datetime.today()
        try:
            os.makedirs(pjoin(path, 'profiles'))
            with open(pjoin(path, 'profiles', 'repo_name'), 'w') as f:
                f.write(f'{repo_id}\n')
            with open(pjoin(path, 'profiles', 'eapi'), 'w') as f:
                f.write(f'{eapi}\n')
            os.makedirs(pjoin(path, 'metadata'))
            with open(pjoin(path, 'metadata', 'layout.conf'), 'w') as f:
                f.write(textwrap.dedent(f"""\
                    masters = {' '.join(masters)}
                    cache-formats =
                    thin-manifests = true
                """))
            if arches:
                self.arches.update(arches)
            os.makedirs(pjoin(path, 'eclass'))
        except FileExistsError:
            pass
        self.sync()

    def sync(self):
        """Forcibly create underlying repo object avoiding cache usage."""
        # avoid issues loading modules that set signal handlers
        from pkgcore.ebuild import repo_objs, repository
        repo_config = repo_objs.RepoConfig(location=self.path, disable_inst_caching=True)
        self._repo = repository.UnconfiguredTree(self.path, repo_config=repo_config)

    def create_profiles(self, profiles):
        for p in profiles:
            os.makedirs(pjoin(self.path, 'profiles', p.path), exist_ok=True)
            with open(pjoin(self.path, 'profiles', 'profiles.desc'), 'a+') as f:
                f.write(f'{p.arch} {p.path} {p.status}\n')
            if p.deprecated:
                with open(pjoin(self.path, 'profiles', p.path, 'deprecated'), 'w') as f:
                    f.write("# deprecated\ndeprecation reason\n")
            with open(pjoin(self.path, 'profiles', p.path, 'make.defaults'), 'w') as f:
                if p.defaults is not None:
                    f.write('\n'.join(p.defaults))
                else:
                    f.write(f'ARCH={p.arch}\n')
            if p.eapi:
                with open(pjoin(self.path, 'profiles', p.path, 'eapi'), 'w') as f:
                    f.write(f'{p.eapi}\n')

    def create_ebuild(self, cpvstr, data=None, **kwargs):
        from pkgcore.ebuild import cpv as cpv_mod
        cpv = cpv_mod.VersionedCPV(cpvstr)
        self._repo.notify_add_package(cpv)
        ebuild_dir = pjoin(self.path, cpv.category, cpv.package)
        os.makedirs(ebuild_dir, exist_ok=True)

        # use defaults for some ebuild metadata if unset
        eapi = kwargs.pop('eapi', '7')
        slot = kwargs.pop('slot', '0')
        desc = kwargs.pop('description', 'stub package description')
        homepage = kwargs.pop('homepage', 'https://github.com/pkgcore/pkgcheck')
        license = kwargs.pop('license', 'blank')

        ebuild_path = pjoin(ebuild_dir, f'{cpv.package}-{cpv.fullver}.ebuild')
        with open(ebuild_path, 'w') as f:
            if self.repo_id == 'gentoo':
                f.write(textwrap.dedent(f"""\
                    # Copyright 1999-{self._today.year} Gentoo Authors
                    # Distributed under the terms of the GNU General Public License v2
                """))
            f.write(f'EAPI="{eapi}"\n')
            f.write(f'DESCRIPTION="{desc}"\n')
            f.write(f'HOMEPAGE="{homepage}"\n')
            f.write(f'SLOT="{slot}"\n')

            if license:
                f.write(f'LICENSE="{license}"\n')
                # create a fake license
                os.makedirs(pjoin(self.path, 'licenses'), exist_ok=True)
                touch(pjoin(self.path, 'licenses', license))

            for k, v in kwargs.items():
                # handle sequences such as KEYWORDS and IUSE
                if isinstance(v, (tuple, list)):
                    v = ' '.join(v)
                f.write(f'{k.upper()}="{v}"\n')
            if data:
                f.write(data.strip() + '\n')

        return ebuild_path

    def __iter__(self):
        yield from iter(self._repo)

    __getattr__ = klass.GetAttrProxy('_repo')
    __dir__ = klass.DirProxy('_repo')


@pytest.fixture
def repo(tmp_path_factory):
    """Create a generic ebuild repository."""
    return EbuildRepo(str(tmp_path_factory.mktemp('repo')))


@pytest.fixture
def make_repo(tmp_path_factory):
    """Factory for ebuild repo creation."""
    def _make_repo(path=None, **kwargs):
        path = str(tmp_path_factory.mktemp('repo')) if path is None else path
        return EbuildRepo(path, **kwargs)
    return _make_repo



__all__ = ("tar_syncer",)

import atexit
import os
import shutil
import subprocess
import tempfile
from functools import partial

from . import base
from .http import http_syncer


class tar_syncer(http_syncer, base.ExternalSyncer):

    binary = 'tar'

    supported_uris = (
        ('tar+http://', 5),
        ('tar+https://', 5),
    )

    # TODO: support more of the less used file extensions
    supported_protocols = ('http://', 'https://')
    supported_exts = ('.tar.gz', '.tar.bz2', '.tar.xz')

    @classmethod
    def parse_uri(cls, raw_uri):
        if raw_uri.startswith(("tar+http://", "tar+https://")):
            raw_uri = raw_uri[4:]
        if raw_uri.startswith(cls.supported_protocols) and raw_uri.endswith(cls.supported_exts):
            return raw_uri
        else:
            raise base.UriError(
                raw_uri, "unsupported compression format for tarball archive")
        raise base.UriError(raw_uri, "unsupported URI")

    def _pre_download(self):
        # create temp file for downloading
        self.tarball = tempfile.NamedTemporaryFile()
        # make sure temp file is deleted on exit
        atexit.register(partial(self.tarball.close))

        # determine names of tempdirs for staging
        basedir = self.basedir.rstrip(os.path.sep)
        repos_dir = os.path.dirname(basedir)
        repo_name = os.path.basename(basedir)
        self.tempdir = os.path.join(repos_dir, f'.{repo_name}.update')
        self.tempdir_old = os.path.join(repos_dir, f'.{repo_name}.old')
        # remove tempdirs on exit
        atexit.register(partial(shutil.rmtree, self.tempdir, ignore_errors=True))
        atexit.register(partial(shutil.rmtree, self.tempdir_old, ignore_errors=True))
        return self.tarball.name

    def _post_download(self, path):
        super()._post_download(path)

        # create tempdirs for staging
        try:
            os.makedirs(self.tempdir)
            os.makedirs(self.tempdir_old)
        except OSError as e:
            raise base.SyncError(f'failed creating repo update dirs: {e}')

        exts = {'gz': 'gzip', 'bz2': 'bzip2', 'xz': 'xz'}
        compression = exts[self.uri.rsplit('.', 1)[1]]
        # use tar instead of tarfile so we can easily strip leading path components
        # TODO: programmatically determine how many components to strip?
        cmd = [
            'tar', '--extract', f'--{compression}', '-f', self.tarball.name,
            '--strip-components=1', '--no-same-owner', '-C', self.tempdir
        ]

        try:
            subprocess.run(cmd, stderr=subprocess.PIPE, check=True, encoding='utf8')
        except subprocess.CalledProcessError as e:
            error = e.stderr.splitlines()[0]
            raise base.SyncError(f'failed to unpack tarball: {error}')

        # TODO: verify gpg data if it exists

        try:
            if os.path.exists(self.basedir):
                # move old repo out of the way if it exists
                os.rename(self.basedir, self.tempdir_old)
            # move new, unpacked repo into place
            os.rename(self.tempdir, self.basedir)
        except OSError as e:
            raise base.SyncError(f'failed to update repo: {e.strerror}') from e



"""repository cache clone utility"""

import time

from ..util import commandline

argparser = commandline.ArgumentParser(
    domain=False, description=__doc__, script=(__file__, __name__))
argparser.add_argument(
    "source", config_type='cache', priority=20,
    action=commandline.StoreConfigObject,
    help="source cache to copy data from")
argparser.add_argument(
    "target", config_type='cache', priority=21,
    action=commandline.StoreConfigObject, writable=True,
    help="target cache to update.  Must be writable.")


@argparser.bind_main_func
def main(options, out, err):
    if options.target.readonly:
        argparser.error(
            "can't update cache label '%s', it's marked readonly." %
            (options.target,))

    source, target = options.source, options.target
    if not target.autocommits:
        target.sync_rate = 1000
    if options.verbosity > 0:
        out.write("grabbing target's existing keys")
    valid = set()
    start = time.time()
    if options.verbosity > 0:
        for k, v in source.items():
            out.write(f"updating {k}")
            target[k] = v
            valid.add(k)
    else:
        for k, v in source.items():
            target[k] = v
            valid.add(k)

    for x in target.keys():
        if x not in valid:
            if options.verbosity > 0:
                out.write(f"deleting {x}")
            del target[x]

    if options.verbosity > 0:
        out.write("took %i seconds" % int(time.time() - start))



"""Helpers for testing scripts."""

from snakeoil.cli import arghparse
from snakeoil.test import argparse_helpers

from ...config import basics, central
from ...config.hint import ConfigHint


class fake_domain:
    pkgcore_config_type = ConfigHint(typename='domain')

    def __init__(self):
        pass

default_domain = basics.HardCodedConfigSection({
    'class': fake_domain,
    'default': True,
    })


class ArgParseMixin(argparse_helpers.ArgParseMixin):
    """Provide some utility methods for testing the parser and main.

    :cvar parser: ArgumentParser subclass to test.
    :cvar main: main function to test.
    """

    requires_compat_config_manager = True
    suppress_domain = False
    has_config = True

    def _mk_config(self, *args, **kwds):
        config = central.ConfigManager(*args, **kwds)
        if self.requires_compat_config_manager:
            config = central.CompatConfigManager(config)
        return config

    def parse(self, *args, **kwargs):
        """Parse a commandline and return the Values object.

        args are passed to parse_args, keyword args are used as config keys.
        """
        ns_kwargs = kwargs.pop('ns_kwargs', {})
        namespace = kwargs.get('namespace', arghparse.Namespace(**ns_kwargs))
        if self.has_config:
            if kwargs.pop("suppress_domain", self.suppress_domain):
                kwargs["default_domain"] = default_domain
            namespace.config = self._mk_config([kwargs], debug=True)
        namespace = self.parser.parse_args(list(args), namespace=namespace)
        return namespace

    def assertOutAndErr(self, *args, **kwargs):
        options = argparse_helpers.ArgParseMixin.assertOutAndErr(self, *args, **kwargs)
        return options.config



__all__ = (
    "SimpleTree", "RepositoryGroup",
    "get_raw_repos", "get_virtual_repos",
)

from snakeoil import klass
from snakeoil.mappings import DictMixin

from ..ebuild.cpv import VersionedCPV
from . import multiplex, prototype, virtual


class SimpleTree(prototype.tree):
    """Fake, in-memory repository.

    Args:
        cpv_dict (dict): CPVs to populate the repo with
        pkg_klass: class of packages in repo
        livefs (bool): regular repo if False, vdb if True
        frozen (bool): repo is modifiable if False, otherwise readonly
        repo_id (str): repo ID
    """

    def __init__(self, cpv_dict, pkg_klass=None, livefs=False, frozen=True,
                 repo_id=None):
        self.cpv_dict = cpv_dict
        if pkg_klass is None:
            pkg_klass = VersionedCPV
        self.livefs = livefs
        self.repo_id = repo_id
        self.package_class = pkg_klass
        super().__init__(frozen=frozen)

    def _get_categories(self, *arg):
        if arg:
            return ()
        return tuple(self.cpv_dict.keys())

    def _get_packages(self, category):
        return tuple(self.cpv_dict[category].keys())

    def _get_versions(self, cp_key):
        return tuple(self.cpv_dict[cp_key[0]][cp_key[1]])

    def notify_remove_package(self, pkg):
        vers = self.cpv_dict[pkg.category][pkg.package]
        vers = [x for x in vers if x != pkg.fullver]
        if vers:
            self.cpv_dict[pkg.category][pkg.package] = vers
        else:
            del self.cpv_dict[pkg.category][pkg.package]
            if not self.cpv_dict[pkg.category]:
                del self.cpv_dict[pkg.category]
        super().notify_remove_package(pkg)

    def notify_add_package(self, pkg):
        self.cpv_dict.setdefault(
            pkg.category, {}).setdefault(pkg.package, []).append(pkg.fullver)
        super().notify_add_package(pkg)


class RepositoryGroup(DictMixin):
    """Group of repos as a single unit.

    Args:
        repos (iterable): repo instances
        combined: combined repo, if None a multiplex repo is created
    """

    __externally_mutable__ = False

    def __init__(self, repos=(), combined=None):
        self.repos = tuple(repos)
        if combined is None:
            combined = multiplex.tree(*self.repos)
        self.combined = combined

    itermatch = klass.alias_attr("combined.itermatch")
    has_match = klass.alias_attr("combined.has_match")
    match = klass.alias_attr("combined.match")
    path_restrict = klass.alias_attr("combined.path_restrict")

    def __contains__(self, key):
        return key in self.combined

    def __iter__(self):
        return iter(self.repos)

    def __getitem__(self, key):
        if isinstance(key, str):
            func = lambda x: key in x.aliases
        elif isinstance(key, int):
            return self.repos[key]
        else:
            func = lambda x: key == x
        try:
            return next(filter(func, self.repos))
        except StopIteration:
            raise KeyError(key)

    def keys(self):
        return (r.repo_id for r in self.repos)

    def items(self):
        return ((r.repo_id, r) for r in self.repos)

    def values(self):
        return iter(self.repos)

    def __add__(self, other):
        if isinstance(other, prototype.tree):
            if other not in self.repos:
                self.repos += (other,)
                self.combined += other
            return self
        elif isinstance(other, RepositoryGroup):
            return RepositoryGroup(self.repos + other.repos)
        elif isinstance(other, (list, tuple)):
            return RepositoryGroup(self.repos + tuple(other))
        raise TypeError(
            "cannot add '%s' and '%s' objects"
            % (self.__class__.__name__, other.__class__.__name__))

    def __radd__(self, other):
        if isinstance(other, prototype.tree):
            if other not in self.repos:
                self.repos = (other,) + self.repos
                self.combined = other + self.combined
            return self
        elif isinstance(other, RepositoryGroup):
            return RepositoryGroup(other.repos + self.repos)
        elif isinstance(other, (list, tuple)):
            return RepositoryGroup(tuple(other) + self.repos)
        raise TypeError(
            "cannot add '%s' and '%s' objects"
            % (other.__class__.__name__, self.__class__.__name__))

    @classmethod
    def change_repos(cls, repos):
        return cls(repos)

    @property
    def real(self):
        return RepositoryGroup(get_virtual_repos(self, False))

    @property
    def virtual(self):
        return RepositoryGroup(get_virtual_repos(self))

    def repo_match(self, path):
        """Find the repo containing a path.

        Args:
            path (str): path in the filesystem

        Returns:
            repo object if a matching repo is found, otherwise None.
        """
        for repo in self.repos:
            if path in repo:
                return repo
        return None


def get_raw_repos(repos):
    """Returns a list of raw repos found.

    repos can be either a repo instance, or a list of repos
    """
    if isinstance(repos, (RepositoryGroup, list, tuple)):
        l = []
        for x in repos:
            l.extend(get_raw_repos(x))
        return l
    while getattr(repos, "raw_repo", None) is not None:
        repos = repos.raw_repo
    if isinstance(repos, multiplex.tree):
        l = []
        for x in repos.trees:
            l.extend(get_raw_repos(x))
        return l
    return [repos]


def get_virtual_repos(repos, sentinel=True):
    """Select only virtual repos.

    repos can be either a list of repos, or a repo to descend through.
    if sentinel is False, will select all non virtual repos
    """
    if not isinstance(repos, (RepositoryGroup, list, tuple)):
        repos = get_raw_repos(repos)
    return [x for x in repos if isinstance(x, (virtual.tree, SimpleTree)) == sentinel]



from pkgcore import gpg


class TestSkipSignatures:

    def test_simple_skipping(self):
        for header in (
                [],
                ["-----BEGIN PGP SIGNED MESSAGE-----\n", "Hash: Sha1\n", "\n"]):
            d = [
                "asdf\n",
                "fdsa\n",
                "-----BEGIN PGP SIGNATURE-----\n",
                "this isn't a valid sig...\n",
                "-----END PGP SIGNATURE-----\n",
                "foon\n"]
            d2 = header + d
            parsed = list(gpg.skip_signatures(d2))
            required = [d[0], d[1], d[-1]]
            assert parsed == required, f"{parsed!r} != {required!r} for header {header!r}"

    def test_signed_signed(self):
        d = [
            "-----BEGIN PGP SIGNED MESSAGE-----\n",
            "Hash: SHA1\n",
            "\n",
            "- -----BEGIN PGP SIGNED MESSAGE-----\n",
            "Hash: SHA1\n",
            "\n",
            "blah\n",
            "- -----BEGIN PGP SIGNATURE-----\n",
            "Version: GnuPG v1.4.3 (GNU/Linux)\n",
            "\n",
            "iD8DBQFEViv+aGfFFLhbXWkRAo+lAJ93s57QA2lW5BE1FdmEc3uzijpJrwCfcE6j",
            "3Nzn/8wExwZ5eUacC/HoSo8=",
            "=oBur",
            "- -----END PGP SIGNATURE-----\n",
            "foon\n",
            "-----BEGIN PGP SIGNATURE-----\n",
            " not valid...\n",
            "-----END PGP SIGNATURE-----\n",
            "asdf\n"]
        assert list(gpg.skip_signatures(d)) == ["blah\n", "foon\n", "asdf\n"]



"""Config introspection support."""


__all__ = ('ConfigHint', 'configurable')


class ConfigHint:
    """Hint for introspection supplying overrides."""

    # be aware this is used in clone
    __slots__ = (
        "types", "positional", "required", "typename", "allow_unknowns",
        "doc", "authorative", "requires_config", "raw_class")

    def __init__(self, types=None, positional=None, required=None, doc=None,
                 typename=None, allow_unknowns=False, authorative=False,
                 requires_config=False, raw_class=False):
        self.types = types or {}
        self.positional = positional or []
        self.required = required or []
        self.typename = typename
        self.allow_unknowns = allow_unknowns
        self.doc = doc
        self.authorative = authorative
        self.requires_config = requires_config
        self.raw_class = raw_class

    def clone(self, **kwds):
        new_kwds = {}
        for attr in self.__slots__:
            new_kwds[attr] = kwds.pop(attr, getattr(self, attr))
        if kwds:
            raise TypeError("unknown type overrides: %r" % kwds)
        return self.__class__(**new_kwds)


def configurable(*args, **kwargs):
    """Decorator version of ConfigHint."""
    hint = ConfigHint(*args, **kwargs)
    def decorator(original):
        original.pkgcore_config_type = hint
        return original
    return decorator



from collections import defaultdict, deque
from operator import itemgetter

from snakeoil.demandload import demand_compile_regexp
from snakeoil.osutils import listdir_files, pjoin
from snakeoil.sequences import iflatten_instance

from ..log import logger
from .atom import atom


def _scan_directory(path, eapi):
    files = []
    for filename in listdir_files(path):
        match = eapi.options.update_regex.match(filename)
        if match is not None:
            files.append(filename)
        else:
            logger.error(f'incorrectly named update file: {filename!r}')
    return sorted(files)


def read_updates(path, eapi):
    def f():
        d = deque()
        return [d,d]
    # mods tracks the start point [0], and the tail, [1].
    # via this, pkg moves into a specific pkg can pick up
    # changes past that point, while ignoring changes prior
    # to that point.
    # Afterwards, we flatten it to get a per cp chain of commands.
    # no need to do lookups basically, although we do need to
    # watch for cycles.
    mods = defaultdict(f)
    moved = {}

    try:
        for fp in _scan_directory(path, eapi):
            with open(pjoin(path, fp)) as f:
                data = (line.rstrip('\n') for line in f)
                _process_updates(data, fp, mods, moved)
    except FileNotFoundError:
        pass

    # force a walk of the tree, flattening it
    commands = {k: list(iflatten_instance(v[0], tuple)) for k,v in mods.items()}
    # filter out empty nodes.
    commands = {k: v for k,v in commands.items() if v}

    return commands


def _process_updates(sequence, filename, mods, moved):
    for lineno, raw_line in enumerate(sequence, 1):
        line = raw_line.strip()
        if not line:
            logger.error(f'file {filename!r}: empty line {lineno}')
            continue
        elif line != raw_line:
            logger.error(
                f'file {filename!r}: extra whitespace in {raw_line!r} on line {lineno}')

        line = line.split()
        if line[0] == 'move':
            if len(line) != 3:
                logger.error(
                    f'file {filename!r}: {raw_line!r} on line {lineno}: bad move form')
                continue
            src, trg = atom(line[1]), atom(line[2])
            if src.fullver is not None:
                logger.error(
                    f"file {filename!r}: {raw_line!r} on line {lineno}: "
                    f"atom {src} must be versionless")
                continue
            elif trg.fullver is not None:
                logger.error(
                    f"file {filename!r}: {raw_line!r} on line {lineno}: "
                    f"atom {trg} must be versionless")
                continue

            if src.key in moved:
                logger.warning(
                    f"file {filename!r}: {raw_line!r} on line {lineno}: "
                    f"{src} was already moved to {moved[src.key]}, "
                    "this line is redundant")
                continue

            d = deque()
            mods[src.key][1].extend([('move', src, trg), d])
            # start essentially a new checkpoint in the trg
            mods[trg.key][1].append(d)
            mods[trg.key][1] = d
            moved[src.key] = trg

        elif line[0] == 'slotmove':
            if len(line) != 4:
                logger.error(
                    f"file {filename!r}: {raw_line!r} on line {lineno}: "
                    "bad slotmove form")
                continue
            src = atom(line[1])

            if src.key in moved:
                logger.warning(
                    f"file {filename!r}: {raw_line!r} on line {lineno}: "
                    f"{src} was already moved to {moved[src.key]}, "
                    "this line is redundant")
                continue
            elif src.slot is not None:
                logger.error(
                    f"file {filename!r}: {raw_line!r} on line {lineno}: "
                    "slotted atom makes no sense for slotmoves")
                continue

            src_slot = atom(f'{src}:{line[2]}')
            trg_slot = atom(f'{src.key}:{line[3]}')

            mods[src.key][1].append(('slotmove', src_slot, line[3]))
        else:
            logger.error(
                f'file {filename!r}: {raw_line!r} on line {lineno}: unknown command')



"""
Internal constants.
"""

import os
import sys

from snakeoil import mappings

from . import __title__

osp = os.path
_reporoot = osp.realpath(__file__).rsplit(os.path.sep, 3)[0]
_module = sys.modules[__name__]

try:
    # This is a file written during pkgcore installation;
    # if it exists, we defer to it.  If it doesn't, then we're
    # running from a git checkout or a tarball.
    from . import _const as _defaults
except ImportError:
    _defaults = object()


def _GET_CONST(attr, default_value):
    consts = mappings.ProxiedAttrs(_module)
    is_tuple = not isinstance(default_value, str)
    if is_tuple:
        default_value = tuple(x % consts for x in default_value)
    else:
        default_value %= consts

    result = getattr(_defaults, attr, default_value)
    if is_tuple:
        result = tuple(result)
    return result


# determine XDG compatible paths
for xdg_var, var_name, fallback_dir in (
        ('XDG_CONFIG_HOME', 'USER_CONFIG_PATH', '~/.config'),
        ('XDG_CACHE_HOME', 'USER_CACHE_PATH', '~/.cache'),
        ('XDG_DATA_HOME', 'USER_DATA_PATH', '~/.local/share')):
    setattr(_module, var_name,
            os.environ.get(xdg_var, osp.join(osp.expanduser(fallback_dir), __title__)))

USER_CONF_FILE = osp.join(getattr(_module, 'USER_CONFIG_PATH'), 'pkgcore.conf')
SYSTEM_CONF_FILE = '/etc/pkgcore/pkgcore.conf'
SYSTEM_CACHE_PATH = '/var/cache/pkgcore'

REPO_PATH = _GET_CONST('REPO_PATH', _reporoot)
DATA_PATH = _GET_CONST('DATA_PATH', '%(REPO_PATH)s/data')
LIBDIR_PATH = _GET_CONST('LIBDIR_PATH', '%(REPO_PATH)s')
CONFIG_PATH = _GET_CONST('CONFIG_PATH', '%(REPO_PATH)s/data/config')
PATH_FORCED_PREPEND = _GET_CONST('INJECTED_BIN_PATH', ('%(REPO_PATH)s/bin',))
CP_BINARY = _GET_CONST('CP_BINARY', '/bin/cp')



"""
Ebuild repository, specific to gentoo ebuild trees.
"""

__all__ = ("UnconfiguredTree", "ConfiguredTree", "ProvidesRepo", "tree")

import locale
import os
from functools import partial, wraps
from itertools import chain, filterfalse
from random import shuffle
from sys import intern
from weakref import WeakValueDictionary

from snakeoil import chksum, klass
from snakeoil.bash import read_dict
from snakeoil.containers import InvertedContains
from snakeoil.data_source import local_source
from snakeoil.mappings import ImmutableDict
from snakeoil.obj import make_kls
from snakeoil.osutils import listdir_dirs, listdir_files, pjoin
from snakeoil.sequences import iflatten_instance, stable_unique
from snakeoil.strings import pluralism

from .. import fetch
from ..config.hint import ConfigHint, configurable
from ..log import logger
from ..operations import repo as _repo_ops
from ..package import errors as pkg_errors
from ..repository import configured, errors, prototype, util
from ..repository.virtual import RestrictionRepo
from ..restrictions import packages
from ..util import packages as pkgutils
from . import cpv, digest, ebd, ebuild_src
from . import eclass_cache as eclass_cache_mod
from . import processor, repo_objs, restricts
from .eapi import get_eapi


class repo_operations(_repo_ops.operations):

    def _cmd_implementation_manifest(self, domain, restriction, observer,
                                     mirrors=False, force=False, distdir=None):
        manifest_config = self.repo.config.manifests
        if manifest_config.disabled:
            observer.info(f'{self.repo.repo_id} repo has manifests disabled')
            return
        required_chksums = set(manifest_config.required_hashes)
        write_chksums = manifest_config.hashes

        if distdir is None:
            distdir = domain.distdir
        ret = set()

        matches = self.repo.itermatch(restriction, sorter=sorted)
        for pkgs in map(list, pkgutils.groupby_pkg(matches)):
            key = pkgs[0].key
            manifest = pkgs[0].manifest

            # check for pkgs masked by bad metadata
            if bad_metadata := self.repo._bad_masked.match(pkgs[0].unversioned_atom):
                for pkg in bad_metadata:
                    e = pkg.data
                    error_str = f"{pkg.cpvstr}: {e.msg(verbosity=observer.verbosity)}"
                    observer.error(error_str)
                    ret.add(pkg.key)
                continue

            # all pkgdir fetchables
            pkgdir_fetchables = {}
            for pkg in pkgs:
                pkgdir_fetchables.update({
                    fetchable.filename: fetchable for fetchable in
                    iflatten_instance(pkg.generate_fetchables(
                        allow_missing_checksums=True,
                        skip_default_mirrors=(not mirrors)),
                        fetch.fetchable)
                })

            # fetchables targeted for (re-)manifest generation
            fetchables = {}
            chksum_set = set(write_chksums)
            for filename, fetchable in pkgdir_fetchables.items():
                if force or not required_chksums.issubset(fetchable.chksums):
                    fetchable.chksums = {
                        k: v for k, v in fetchable.chksums.items() if k in chksum_set}
                    fetchables[filename] = fetchable

            # Manifest files aren't necessary with thin manifests and no distfiles
            if manifest_config.thin and not pkgdir_fetchables:
                if os.path.exists(manifest.path):
                    try:
                        os.remove(manifest.path)
                    except EnvironmentError as e:
                        observer.error(
                            'failed removing old manifest: '
                            f'{key}::{self.repo.repo_id}: {e}')
                        ret.add(key)
                continue

            # Manifest file is current and not forcing a refresh
            if not force and manifest.distfiles.keys() == pkgdir_fetchables.keys():
                continue

            # fetch distfiles
            pkg_ops = domain.pkg_operations(pkg, observer=observer)
            if not pkg_ops.fetch(list(fetchables.values()), observer, distdir=distdir):
                ret.add(key)
                continue

            # calculate checksums for fetched distfiles
            try:
                for fetchable in fetchables.values():
                    chksums = chksum.get_chksums(
                        pjoin(distdir, fetchable.filename), *write_chksums)
                    fetchable.chksums = dict(zip(write_chksums, chksums))
            except chksum.MissingChksumHandler as e:
                observer.error(f'failed generating chksum: {e}')
                ret.add(key)
                break

            if key not in ret:
                fetchables.update(pkgdir_fetchables)
                observer.info(f"generating manifest: {key}::{self.repo.repo_id}")
                manifest.update(sorted(fetchables.values()), chfs=write_chksums)

        return ret


def _sort_eclasses(config, repo_config):
    repo_path = repo_config.location
    masters = repo_config.masters
    eclasses = []

    default = config.get_default('repo_config')
    if repo_config._missing_masters and default is not None:
        # use default repo's eclasses for overlays with missing masters
        location = default.location
    else:
        location = repo_path

    if not masters:
        eclasses = [location]
    else:
        repo_map = {
            alias: r.location for r in config.objects['repo_config'].values()
            for alias in r.aliases}
        eclasses = [repo_map[x] for x in masters]

    # add the repo's eclass directories if it's not specified.
    # do it in this fashion so that the repo's masters can actually interpose
    # this repo's eclasses in between others.
    # admittedly an odd thing to do, but it has some benefits
    if repo_path not in eclasses:
        eclasses.append(repo_path)

    eclasses = [eclass_cache_mod.cache(pjoin(x, 'eclass'), location=location)
                for x in eclasses]

    if len(eclasses) == 1:
        eclasses = eclasses[0]
    else:
        eclasses = list(reversed(eclasses))
        eclasses = eclass_cache_mod.StackedCaches(
            eclasses, location=location, eclassdir=location)
    return eclasses


class ProvidesRepo(util.SimpleTree):
    """Fake, installed repo populated with entries from package.provided."""

    class PkgProvidedParent:

        def __init__(self, **kwds):
            self.__dict__.update(kwds)

    class PkgProvided(ebuild_src.base):

        __slots__ = ('arches', 'use')

        package_is_real = False
        __inst_caching__ = True

        @property
        def keywords(self):
            return self.arches

        def __init__(self, *args, arches=(), **kwargs):
            super().__init__(*args, **kwargs)
            object.__setattr__(self, "use", [])
            object.__setattr__(self, "arches", arches)
            object.__setattr__(self, "data", {"SLOT": "0"})
            object.__setattr__(self, "eapi", get_eapi('0'))

    def __init__(self, pkgs, arches, repo_id='package.provided'):
        d = {}
        for pkg in pkgs:
            d.setdefault(pkg.category, {}).setdefault(pkg.package, []).append(pkg.fullver)
        intermediate_parent = self.PkgProvidedParent()
        super().__init__(
            d, pkg_klass=partial(self.PkgProvided, intermediate_parent, arches=arches),
            livefs=True, frozen=True, repo_id=repo_id)
        intermediate_parent._parent_repo = self

        if not d:
            self.match = self.itermatch = self._empty_provides_iterable
            self.has_match = self._empty_provides_has_match

    @staticmethod
    def _empty_provides_iterable(*args, **kwds):
        return iter(())

    @staticmethod
    def _empty_provides_has_match(*args, **kwds):
        return False


class UnconfiguredTree(prototype.tree):
    """Raw implementation supporting standard ebuild tree.

    Return packages don't have USE configuration bound to them.
    """

    false_categories = frozenset(["eclass", "profiles", "metadata", "licenses"])
    configured = False
    configurables = ("domain", "settings")
    package_factory = staticmethod(ebuild_src.generate_new_factory)
    enable_gpg = False
    extension = '.ebuild'

    operations_kls = repo_operations

    pkgcore_config_type = ConfigHint({
        'location': 'str',
        'eclass_cache': 'ref:eclass_cache',
        'masters': 'refs:repo',
        'cache': 'refs:cache',
        'default_mirrors': 'list',
        'allow_missing_manifests': 'bool',
        'repo_config': 'ref:repo_config',
        },
        typename='repo')

    def __init__(self, location, eclass_cache=None, masters=(), cache=(),
                 default_mirrors=None, allow_missing_manifests=False, package_cache=True,
                 repo_config=None):
        """
        :param location: on disk location of the tree
        :param cache: sequence of :obj:`pkgcore.cache.template.database` instances
            to use for storing metadata
        :param masters: repo masters this repo inherits from
        :param eclass_cache: If not None, :obj:`pkgcore.ebuild.eclass_cache`
            instance representing the eclasses available,
            if None, generates the eclass_cache itself
        :param default_mirrors: Either None, or sequence of mirrors to try
            fetching from first, then falling back to other uri
        :param package_cache: boolean controlling package instance caching
        :param repo_config: :obj:`pkgcore.repo_objs.RepoConfig` instance for the related repo
        """
        super().__init__()
        self.base = self.location = location
        self.package_cache = package_cache
        if repo_config is None:
            repo_config = repo_objs.RepoConfig(location)
        self.config = repo_config

        # profiles dir is required by PMS
        if not os.path.isdir(self.config.profiles_base):
            raise errors.InvalidRepo(f'missing required profiles dir: {self.location!r}')

        # verify we support the repo's EAPI
        if not self.is_supported:
            raise errors.UnsupportedRepo(self)

        if eclass_cache is None:
            eclass_cache = eclass_cache_mod.cache(
                pjoin(self.location, 'eclass'), location=self.location)
        self.eclass_cache = eclass_cache

        self.masters = tuple(masters)
        self.trees = self.masters + (self,)
        self.licenses = repo_objs.Licenses(self, *self.masters)
        self.profiles = self.config.profiles
        if masters:
            self.profiles = repo_objs.OverlayedProfiles(*self.trees)

        # use mirrors from masters if not defined in the repo
        mirrors = dict(self.thirdpartymirrors)
        for master in masters:
            for k, v in master.mirrors.items():
                if k not in mirrors:
                    mirrors[k] = v

        if isinstance(cache, (tuple, list)):
            cache = tuple(cache)
        else:
            cache = (cache,)

        self.mirrors = mirrors
        self.default_mirrors = default_mirrors
        self.cache = cache
        self._allow_missing_chksums = allow_missing_manifests
        self.package_class = self.package_factory(
            self, cache, self.eclass_cache, self.mirrors, self.default_mirrors)
        self._shared_pkg_cache = WeakValueDictionary()
        self._bad_masked = RestrictionRepo(repo_id='bad_masked')
        self.projects_xml = repo_objs.LocalProjectsXml(
            pjoin(self.location, 'metadata', 'projects.xml'))

    repo_id = klass.alias_attr('config.repo_id')
    repo_name = klass.alias_attr('config.repo_name')
    aliases = klass.alias_attr('config.aliases')
    eapi = klass.alias_attr('config.eapi')
    is_supported = klass.alias_attr('config.eapi.is_supported')
    external = klass.alias_attr('config.external')
    pkg_masks = klass.alias_attr('config.pkg_masks')

    def configure(self, *args):
        return ConfiguredTree(self, *args)

    @klass.jit_attr
    def known_arches(self):
        """Return all known arches for a repo (including masters)."""
        return frozenset(chain.from_iterable(
            r.config.known_arches for r in self.trees))

    def path_restrict(self, path):
        """Return a restriction from a given path in a repo.

        :param path: full or partial path to an ebuild
        :return: a package restriction matching the given path if possible
        :raises ValueError: if the repo doesn't contain the given path, the
            path relates to a file that isn't an ebuild, or the ebuild isn't in the
            proper directory layout
        """
        if path not in self:
            raise ValueError(f"{self.repo_id!r} repo doesn't contain: {path!r}")

        if not path.startswith(os.sep) and os.path.exists(pjoin(self.location, path)):
            path_chunks = path.split(os.path.sep)
        else:
            path = os.path.realpath(os.path.abspath(path))
            relpath = path[len(os.path.realpath(self.location)):].strip('/')
            path_chunks = relpath.split(os.path.sep)

        if os.path.isfile(path):
            if not path.endswith('.ebuild'):
                raise ValueError(f"file is not an ebuild: {path!r}")
            elif len(path_chunks) != 3:
                # ebuild isn't in a category/PN directory
                raise ValueError(
                    f"ebuild not in the correct directory layout: {path!r}")

        restrictions = []

        # add restrictions until path components run out
        try:
            restrictions.append(restricts.RepositoryDep(self.repo_id))
            if path_chunks[0] in self.categories:
                restrictions.append(restricts.CategoryDep(path_chunks[0]))
                restrictions.append(restricts.PackageDep(path_chunks[1]))
                base = cpv.VersionedCPV(f"{path_chunks[0]}/{os.path.splitext(path_chunks[2])[0]}")
                restrictions.append(restricts.VersionMatch('=', base.version, rev=base.revision))
        except IndexError:
            pass
        return packages.AndRestriction(*restrictions)

    def __getitem__(self, cpv):
        cpv_inst = self.package_class(*cpv)
        if cpv_inst.fullver not in self.versions[(cpv_inst.category, cpv_inst.package)]:
            raise KeyError(cpv)
        return cpv_inst

    def rebind(self, **kwds):
        """Generate a new tree instance with the same location using new keywords.

        :param kwds: see __init__ for valid values
        """
        o = self.__class__(self.location, **kwds)
        o.categories = self.categories
        o.packages = self.packages
        o.versions = self.versions
        return o

    @klass.jit_attr
    def thirdpartymirrors(self):
        mirrors = {}
        fp = pjoin(self.location, 'profiles', 'thirdpartymirrors')
        try:
            for k, v in read_dict(fp, splitter=None).items():
                v = v.split()
                # shuffle mirrors so the same ones aren't used every time
                shuffle(v)
                mirrors[k] = v
        except FileNotFoundError:
            pass
        return ImmutableDict(mirrors)

    @klass.jit_attr
    def use_expand_desc(self):
        """Inherited USE_EXPAND settings for the repo."""
        d = {}
        for repo in self.trees:
            d.update(repo.config.use_expand_desc)
        return ImmutableDict(d)

    @klass.jit_attr
    def use_expand_sort(self):
        """Inherited mapping of USE_EXPAND sorting keys for the repo."""
        d = {}
        for repo in self.trees:
            d.update(repo.config.use_expand_sort)
        return ImmutableDict(d)

    def use_expand_sorter(self, group):
        """Sorting function for a given USE_EXPAND group."""
        try:
            ordering = self.use_expand_sort[group.lower()]
            return lambda k: ordering.get(k, -1)
        except KeyError:
            # nonexistent USE_EXPAND group
            return lambda k: k

    @klass.jit_attr
    def category_dirs(self):
        try:
            return frozenset(map(intern, filterfalse(
                self.false_categories.__contains__,
                (x for x in listdir_dirs(self.base) if not x.startswith('.')))))
        except EnvironmentError as e:
            logger.error(f"failed listing categories: {e}")
        return ()

    def _get_categories(self, *optional_category):
        # why the auto return? current porttrees don't allow/support
        # categories deeper then one dir.
        if optional_category:
            # raise KeyError
            return ()
        categories = frozenset(chain.from_iterable(repo.config.categories for repo in self.trees))
        if categories:
            return categories
        return self.category_dirs

    def _get_packages(self, category):
        cpath = pjoin(self.base, category.lstrip(os.path.sep))
        try:
            return tuple(listdir_dirs(cpath))
        except FileNotFoundError:
            if category in self.categories:
                # ignore it, since it's PMS mandated that it be allowed.
                return ()
        except EnvironmentError as e:
            category = pjoin(self.base, category.lstrip(os.path.sep))
            raise KeyError(
                f'failed fetching packages for category {category}: {e}') from e

    def _get_versions(self, catpkg):
        """Determine available versions for a given package.

        Ebuilds with mismatched or invalid package names are ignored.
        """
        cppath = pjoin(self.base, catpkg[0], catpkg[1])
        pkg = f'{catpkg[-1]}-'
        lp = len(pkg)
        extension = self.extension
        ext_len = -len(extension)
        try:
            return tuple(
                x[lp:ext_len] for x in listdir_files(cppath)
                if x[ext_len:] == extension and x[:lp] == pkg)
        except EnvironmentError as e:
            raise KeyError(
                "failed fetching versions for package %s: %s" %
                (pjoin(self.base, '/'.join(catpkg)), str(e))) from e

    def _pkg_filter(self, raw, error_callback, pkgs):
        """Filter packages with bad metadata."""
        while True:
            try:
                pkg = next(pkgs)
            except pkg_errors.PackageError:
                # ignore pkgs with invalid CPVs
                continue
            except StopIteration:
                return

            if raw:
                yield pkg
            elif self._bad_masked.has_match(pkg.versioned_atom) and error_callback is not None:
                error_callback(self._bad_masked[pkg.versioned_atom])
            else:
                # check pkgs for unsupported/invalid EAPIs and bad metadata
                try:
                    if not pkg.is_supported:
                        exc = pkg_errors.MetadataException(
                            pkg, 'eapi', f"EAPI '{pkg.eapi}' is not supported")
                        self._bad_masked[pkg.versioned_atom] = exc
                        if error_callback is not None:
                            error_callback(exc)
                        continue
                    # TODO: add a generic metadata validation method to avoid slow metadata checks?
                    pkg.data
                    pkg.slot
                    pkg.required_use
                except pkg_errors.MetadataException as e:
                    self._bad_masked[e.pkg.versioned_atom] = e
                    if error_callback is not None:
                        error_callback(e)
                    continue
                yield pkg

    def itermatch(self, *args, **kwargs):
        raw = 'raw_pkg_cls' in kwargs or not kwargs.get('versioned', True)
        error_callback = kwargs.pop('error_callback', None)
        kwargs.setdefault('pkg_filter', partial(self._pkg_filter, raw, error_callback))
        return super().itermatch(*args, **kwargs)

    def _get_ebuild_path(self, pkg):
        return pjoin(
            self.base, pkg.category, pkg.package,
            f"{pkg.package}-{pkg.fullver}{self.extension}")

    def _get_ebuild_src(self, pkg):
        return local_source(self._get_ebuild_path(pkg), encoding='utf8')

    def _get_shared_pkg_data(self, category, package):
        key = (category, package)
        o = self._shared_pkg_cache.get(key)
        if o is None:
            mxml = self._get_metadata_xml(category, package)
            manifest = self._get_manifest(category, package)
            o = repo_objs.SharedPkgData(mxml, manifest)
            self._shared_pkg_cache[key] = o
        return o

    def _get_metadata_xml(self, category, package):
        return repo_objs.LocalMetadataXml(pjoin(
            self.base, category, package, "metadata.xml"))

    def _get_manifest(self, category, package):
        return digest.Manifest(pjoin(
            self.base, category, package, "Manifest"),
            thin=self.config.manifests.thin,
            enforce_gpg=self.enable_gpg)

    def _get_digests(self, pkg, allow_missing=False):
        if self.config.manifests.disabled:
            return True, {}
        try:
            manifest = pkg._shared_pkg_data.manifest
            manifest.allow_missing = allow_missing
            return allow_missing, manifest.distfiles
        except pkg_errors.ParseChksumError as e:
            if e.missing and allow_missing:
                return allow_missing, {}
            raise pkg_errors.MetadataException(pkg, 'manifest', str(e))

    def __repr__(self):
        return "<ebuild %s location=%r @%#8x>" % (
            self.__class__.__name__, self.base, id(self))

    @klass.jit_attr
    def deprecated(self):
        """Base deprecated packages restriction from profiles/package.deprecated."""
        pkg_deprecated = set()
        for repo in self.trees:
            pkg_deprecated.update(repo.config.pkg_deprecated)
        return packages.OrRestriction(*pkg_deprecated)

    def _regen_operation_helper(self, **kwds):
        return _RegenOpHelper(
            self, force=bool(kwds.get('force', False)),
            eclass_caching=bool(kwds.get('eclass_caching', True)))

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_shared_pkg_cache']
        return d

    def __setstate__(self, state):
        self.__dict__ = state.copy()
        self.__dict__['_shared_pkg_cache'] = WeakValueDictionary()


@configurable(
    typename='repo',
    types={
        'repo_config': 'ref:repo_config', 'cache': 'refs:cache',
        'eclass_cache': 'ref:eclass_cache',
        'default_mirrors': 'list',
        'allow_missing_manifests': 'bool'},
    requires_config='config')
def tree(config, repo_config, cache=(), eclass_cache=None,
         default_mirrors=None, allow_missing_manifests=False,
         tree_cls=UnconfiguredTree):
    """Initialize an unconfigured ebuild repository."""
    repo_id = repo_config.repo_id
    repo_path = repo_config.location

    if repo_config.masters is None:
        # if it's None, that means it's not a standalone, and is PMS, or misconfigured.
        # empty tuple means it's a standalone repository
        default = config.get_default('repo_config')
        if default is None:
            raise errors.InitializationError(
                f"repo {repo_id!r} at {repo_path!r} requires missing default repo")

    # map external repo ids to their config names
    config_map = {
        r.repo_id: r.location for r in config.objects['repo_config'].values() if r.external}

    try:
        masters = []
        missing = []
        for r in repo_config.masters:
            if repo := config.objects['repo'].get(config_map.get(r, r)):
                masters.append(repo)
            else:
                missing.append(r)
    except RecursionError:
        repo_id = repo_config.repo_id
        masters = ', '.join(repo_config.masters)
        raise errors.InitializationError(
            f'{repo_id!r} repo has cyclic masters: {masters}')

    if missing:
        missing = ', '.join(map(repr, sorted(missing)))
        raise errors.InitializationError(
            f'repo {repo_id!r} at path {repo_path!r} has missing masters: {missing}')

    if eclass_cache is None:
        eclass_cache = _sort_eclasses(config, repo_config)

    return tree_cls(
        repo_config.location, eclass_cache=eclass_cache, masters=masters, cache=cache,
        default_mirrors=default_mirrors,
        allow_missing_manifests=allow_missing_manifests,
        repo_config=repo_config)


class _RegenOpHelper:

    def __init__(self, repo, force=False, eclass_caching=True):
        self.force = force
        self.eclass_caching = eclass_caching
        self.ebp = self.request_ebp()

    def request_ebp(self):
        ebp = processor.request_ebuild_processor()
        if self.eclass_caching:
            ebp.allow_eclass_caching()
        return ebp

    def __call__(self, pkg):
        try:
            return pkg._fetch_metadata(ebp=self.ebp, force_regen=self.force)
        except pkg_errors.MetadataException as e:
            # ebuild processor is dead, so force a replacement request
            self.ebp = self.request_ebp()
            raise

    def __del__(self):
        if self.eclass_caching:
            self.ebp.disable_eclass_caching()
        processor.release_ebuild_processor(self.ebp)


class ConfiguredTree(configured.tree):
    """Wrapper around a :obj:`UnconfiguredTree` binding build/configuration data (USE)."""

    configurable = "use"
    config_wrappables = {
        x: klass.alias_method("evaluate_depset")
        for x in (
            "bdepend", "depend", "rdepend", "pdepend", "idepend",
            "fetchables", "license", "src_uri", "restrict", "required_use",
        )
    }

    def __init__(self, raw_repo, domain, domain_settings):
        """
        :param raw_repo: :obj:`UnconfiguredTree` instance
        :param domain_settings: environment settings to bind
        """
        required_settings = {'USE', 'CHOST'}
        if missing_settings := required_settings.difference(domain_settings):
            s = pluralism(missing_settings)
            raise errors.InitializationError(
                f"{self.__class__} missing required setting{s}: "
                f"{', '.join(map(repr, missing_settings))}")

        chost = domain_settings['CHOST']
        scope_update = {'chost': chost}
        scope_update.update(
            (x, domain_settings.get(x.upper(), chost))
            for x in ('cbuild', 'ctarget'))
        scope_update.update(
            (x, domain_settings.get(x.upper(), ''))
            for x in ('cflags', 'cxxflags', 'ldflags'))
        scope_update['operations_callback'] = self._generate_pkg_operations

        # update wrapped attr funcs requiring access to the class instance
        for k, v in self.config_wrappables.items():
            if isinstance(v, str):
                self.config_wrappables[k] = getattr(self, v)

        super().__init__(
            raw_repo, self.config_wrappables, pkg_kls_injections=scope_update)

        self.domain = domain
        self.domain_settings = domain_settings
        self._delayed_iuse = partial(make_kls(InvertedContains), InvertedContains)

    def _wrap_attr(config_wrappables):
        """Register wrapped attrs that require class instance access."""
        def _wrap_func(func):
            @wraps(func)
            def wrapped(*args, **kwargs):
                return func(*args, **kwargs)
            attr = func.__name__.lstrip('_')
            config_wrappables[attr] = func.__name__
            return wrapped
        return _wrap_func

    @_wrap_attr(config_wrappables)
    def _iuse_effective(self, raw_pkg_iuse_effective, _enabled_use, pkg):
        """IUSE_EFFECTIVE for a package."""
        profile_iuse_effective = self.domain.profile.iuse_effective
        return frozenset(profile_iuse_effective.union(raw_pkg_iuse_effective))

    @_wrap_attr(config_wrappables)
    def _distfiles(self, raw_pkg_distfiles, enabled_use, pkg):
        """Distfiles used by a package."""
        return tuple(stable_unique(raw_pkg_distfiles.evaluate_depset(enabled_use)))

    @_wrap_attr(config_wrappables)
    def _user_patches(self, _raw_pkg_patches, _enabled_use, pkg):
        """User patches that will be applied when building a package."""
        # determine available user patches for >= EAPI 6
        if pkg.eapi.options.user_patches:
            patches = []
            patchroot = pjoin(self.domain.config_dir, 'patches')
            patch_dirs = [
                pkg.PF,
                f'{pkg.PF}:{pkg.slot}',
                pkg.P,
                f'{pkg.P}:{pkg.slot}',
                pkg.PN,
                f'{pkg.PN}:{pkg.slot}',
            ]
            for d in patch_dirs:
                for root, _dirs, files in os.walk(pjoin(patchroot, pkg.category, d)):
                    files = (
                        pjoin(root, f) for f in sorted(files, key=locale.strxfrm)
                        if f.endswith(('.diff', '.patch')))
                    patches.append((root, tuple(files)))
            return tuple(patches)
        return None

    def _get_delayed_immutable(self, pkg, immutable):
        return InvertedContains(set(pkg.iuse).difference(immutable))

    def _get_pkg_kwds(self, pkg):
        immutable, enabled, _disabled = self.domain.get_package_use_unconfigured(pkg)
        return {
            "initial_settings": enabled,
            "unchangable_settings": self._delayed_iuse(
                self._get_delayed_immutable, pkg, immutable)}

    def _generate_pkg_operations(self, domain, pkg, **kwds):
        return ebd.src_operations(domain, pkg, pkg.repo.eclass_cache, **kwds)



import binascii
import os
import shutil
import stat
import textwrap
from tempfile import NamedTemporaryFile

from pkgcore import const
from pkgcore import exceptions as base_errors
from pkgcore.config import errors as config_errors
from pkgcore.ebuild.portage_conf import PortageConfig
from snakeoil.osutils import pjoin
from snakeoil.test import TestCase
from snakeoil.test.mixins import TempDirMixin

load_make_conf = PortageConfig.load_make_conf
load_repos_conf = PortageConfig.load_repos_conf


class TestPortageConfig(TempDirMixin, TestCase):

    def __init__(self, *args, **kwargs):
        TempDirMixin.__init__(self, *args, **kwargs)
        TestCase.__init__(self, *args, **kwargs)

        # default files
        self.make_globals = {}
        load_make_conf(
            self.make_globals, pjoin(const.CONFIG_PATH, 'make.globals'))

        self.global_repos_defaults, self.global_repos_conf = load_repos_conf(
            pjoin(const.CONFIG_PATH, 'repos.conf'))

    def test_load_make_conf(self):
        self.assertIn('PORTAGE_TMPDIR', self.make_globals)

        # nonexistent file
        d = {}
        # by default files are required
        self.assertRaises(
            config_errors.ParsingError, load_make_conf,
            d, pjoin(self.dir, 'make.globals'))
        # should return empty dict when not required
        load_make_conf(d, pjoin(self.dir, 'make.conf'), required=False)
        self.assertEqual({}, d)

        # unreadable file
        if os.getuid() != 0:
            d = {}
            with NamedTemporaryFile() as f:
                os.chmod(f.name, stat.S_IWUSR)
                self.assertRaises(
                    base_errors.PermissionDenied, load_make_conf, d, f.name)

        # overrides and incrementals
        with NamedTemporaryFile() as f:
            f.write(b'DISTDIR=foo\n')
            f.flush()
            d = {}
            load_make_conf(d, pjoin(const.CONFIG_PATH, 'make.globals'))
            load_make_conf(d, f.name, allow_sourcing=True, incrementals=True)
            self.assertEqual('foo', d['DISTDIR'])

    def test_load_make_conf_dir(self):
        # load files from dir and symlinked dir

        make_conf_dir = pjoin(self.dir, 'make.conf')
        os.mkdir(make_conf_dir)
        make_conf_sym = pjoin(self.dir, 'make.conf.sym')
        os.symlink(make_conf_dir, make_conf_sym)

        with open(pjoin(make_conf_dir, 'a'), 'w') as f:
            f.write('DISTDIR=foo\n')
            f.flush()

            d = {}
            load_make_conf(d, pjoin(const.CONFIG_PATH, 'make.globals'))
            sym_d = d.copy()
            load_make_conf(d, make_conf_dir)
            load_make_conf(sym_d, make_conf_sym)

            self.assertEqual(d, sym_d)
            self.assertEqual('foo', d['DISTDIR'])

    # TODO: add some tests for duplicate sections that should output log messages
    def test_load_repos_conf(self):
        self.assertIn('gentoo', self.global_repos_conf)

        # nonexistent file
        self.assertRaises(
            config_errors.ParsingError, load_repos_conf,
            pjoin(self.dir, 'repos.conf'))

        # unreadable file
        if os.getuid() != 0:
            with NamedTemporaryFile() as f:
                os.chmod(f.name, stat.S_IWUSR)
                self.assertRaises(
                    base_errors.PermissionDenied, load_repos_conf, f.name)

        # blank file
        with NamedTemporaryFile() as f:
            self.assertRaises(
                config_errors.ConfigurationError, load_repos_conf, f.name)

        # garbage file
        with NamedTemporaryFile() as f:
            f.write(binascii.b2a_hex(os.urandom(10)))
            f.flush()
            self.assertRaises(
                config_errors.ConfigurationError, load_repos_conf, f.name)

        # missing location parameter
        with NamedTemporaryFile() as f:
            f.write(textwrap.dedent('''\
                [foo]
                sync-uri = git://foo.git''').encode())
            f.flush()
            self.assertRaises(
                config_errors.ConfigurationError, load_repos_conf, f.name)

        # bad priority value causes fallback to the default
        with NamedTemporaryFile() as f:
            f.write(textwrap.dedent('''\
                [foo]
                priority = foo
                location = /var/gentoo/repos/foo
                [gentoo]
                location = /var/gentoo/repos/gentoo''').encode())
            f.flush()
            defaults, repos = load_repos_conf(f.name)
            self.assertEqual(0, repos['foo']['priority'])

        # overriding defaults in the same file throws an exception from configparser
        with NamedTemporaryFile() as f:
            f.write(textwrap.dedent('''\
                [DEFAULT]
                main-repo = gentoo
                [DEFAULT]
                main-repo = foo

                [foo]
                priority = foo
                location = /var/gentoo/repos/foo
                [gentoo]
                location = /var/gentoo/repos/gentoo''').encode())
            f.flush()
            self.assertRaises(
                config_errors.ConfigurationError, load_repos_conf, f.name)

        # undefined main repo with 'gentoo' missing
        with NamedTemporaryFile() as f:
            f.write(textwrap.dedent('''\
                [foo]
                location = /var/gentoo/repos/foo''').encode())
            f.flush()
            self.assertRaises(
                config_errors.UserConfigError, load_repos_conf, f.name)

        # default section isn't required as long as gentoo repo exists
        with NamedTemporaryFile() as f:
            f.write(textwrap.dedent('''\
                [foo]
                location = /var/gentoo/repos/foo
                [gentoo]
                location = /var/gentoo/repos/gentoo''').encode())
            f.flush()
            defaults, repos = load_repos_conf(f.name)
            self.assertEqual('gentoo', defaults['main-repo'])
            self.assertEqual(['foo', 'gentoo'], list(repos.keys()))

        # TODO: check for logger output?
        # overriding defaults in the same file throws an exception from configparser
        with NamedTemporaryFile() as f:
            f.write(textwrap.dedent('''\
                [DEFAULT]
                main-repo = foo
                [foo]
                priority = 3
                location = /var/gentoo/repos/gentoo
                [foo]
                location = /var/gentoo/repos/foo''').encode())
            f.flush()
            self.assertRaises(
                config_errors.ConfigurationError, load_repos_conf, f.name)

    def test_load_repos_conf_dir(self):
        # repo priority sorting and dir/symlink scanning

        repos_conf_dir = pjoin(self.dir, 'repos.conf')
        os.mkdir(repos_conf_dir)
        repos_conf_sym = pjoin(self.dir, 'repos.conf.sym')
        os.symlink(repos_conf_dir, repos_conf_sym)

        # add global repos.conf
        shutil.copyfile(
            pjoin(const.CONFIG_PATH, 'repos.conf'),
            pjoin(repos_conf_dir, 'repos.conf'))

        with open(pjoin(repos_conf_dir, 'z'), 'w') as f:
            f.write(textwrap.dedent('''\
                [bar]
                location = /var/gentoo/repos/bar

                [foo]
                location = /var/gentoo/repos/foo
                priority = 10'''))
            f.flush()

        defaults, repos = load_repos_conf(repos_conf_dir)
        sym_defaults, sym_repos = load_repos_conf(repos_conf_sym)

        self.assertEqual(defaults, sym_defaults)
        self.assertEqual(repos, sym_repos)
        self.assertEqual('gentoo', defaults['main-repo'])
        self.assertEqual(['foo', 'bar', 'gentoo', 'binpkgs'], list(repos.keys()))



"""gentoo ebuild specific base package class"""

from collections import UserString

from snakeoil.compatibility import cmp
from snakeoil.demandload import demand_compile_regexp

from ..package import base
from . import atom
from .errors import InvalidCPV

demand_compile_regexp(
    'suffix_regexp', '^(alpha|beta|rc|pre|p)(\\d*)$')

suffix_value = {"pre": -2, "p": 1, "alpha": -4, "beta": -3, "rc": -1}

# while the package section looks fugly, there is a reason for it-
# to prevent version chunks from showing up in the package

demand_compile_regexp(
    'isvalid_version_re',
    r"^(?:\d+)(?:\.\d+)*[a-zA-Z]?(?:_(p(?:re)?|beta|alpha|rc)\d*)*$")

demand_compile_regexp(
    'isvalid_cat_re', r"^(?:[a-zA-Z0-9][-a-zA-Z0-9+._]*(?:/(?!$))?)+$")

# empty string is fine, means a -- was encounter.
demand_compile_regexp(
    '_pkg_re', r"^[a-zA-Z0-9+_]+$")


def isvalid_pkg_name(chunks):
    if not chunks[0] or chunks[0][0] == '+':
        # this means a leading -; additionally, '+asdf' is disallowed
        return False
    # all remaining chunks can either be empty (meaning multiple
    # hyphens) or must be valid chars
    if not all(not s or _pkg_re.match(s) for s in chunks):
        return False
    # the package name must not end with a hyphen followed by anything that
    # looks like a version -- need to ensure that we've gotten more than one
    # chunk, i.e. at least one hyphen
    if len(chunks) > 1 and isvalid_version_re.match(chunks[-1]):
        return False
    return True


def isvalid_rev(s):
    return s and s[0] == 'r' and s[1:].isdigit()


class Revision(UserString):
    """Internal revision class storing revisions as strings and comparing as integers."""

    # parent __hash__() isn't inherited when __eq__() is defined in the child class
    # https://docs.python.org/3/reference/datamodel.html#object.__hash__
    __hash__ = UserString.__hash__

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.data:
            try:
                self._revint = int(self.data)
            except ValueError:
                raise InvalidCPV(self.data, 'invalid revision')
        else:
            self._revint = 0

    def __str__(self):
        if not self.data:
            return '0'
        else:
            return self.data

    def __eq__(self, other):
        if isinstance(other, Revision):
            return self._revint == other._revint
        elif isinstance(other, int):
            return self._revint == other
        elif other is None:
            return self._revint == 0
        return self.data == other

    def __lt__(self, other):
        if isinstance(other, Revision):
            return self._revint < other._revint
        elif isinstance(other, int):
            return self._revint < other
        elif other is None:
            return self._revint < 0
        return self.data < other

    def __le__(self, other):
        if isinstance(other, Revision):
            return self._revint <= other._revint
        elif isinstance(other, int):
            return self._revint <= other
        elif other is None:
            return self._revint <= 0
        return self.data <= other

    def __gt__(self, other):
        if isinstance(other, Revision):
            return self._revint > other._revint
        elif isinstance(other, int):
            return self._revint > other
        elif other is None:
            return self._revint > 0
        return self.data > other

    def __ge__(self, other):
        if isinstance(other, Revision):
            return self._revint >= other._revint
        elif isinstance(other, int):
            return self._revint >= other
        elif other is None:
            return self._revint >= 0
        return self.data >= other


def ver_cmp(ver1, rev1, ver2, rev2):
    # If the versions are the same, comparing revisions will suffice.
    if ver1 == ver2:
        # revisions are equal if 0 or None (versionless cpv)
        if not rev1 and not rev2:
            return 0
        return cmp(rev1, rev2)

    # Split up the versions into dotted strings and lists of suffixes.
    parts1 = ver1.split("_")
    parts2 = ver2.split("_")

    # If the dotted strings are equal, we can skip doing a detailed comparison.
    if parts1[0] != parts2[0]:

        # First split up the dotted strings into their components.
        ver_parts1 = parts1[0].split(".")
        ver_parts2 = parts2[0].split(".")

        # Pull out any letter suffix on the final components and keep
        # them for later.
        letters = []
        for ver_parts in (ver_parts1, ver_parts2):
            if ver_parts[-1][-1].isalpha():
                letters.append(ord(ver_parts[-1][-1]))
                ver_parts[-1] = ver_parts[-1][:-1]
            else:
                # Using -1 simplifies comparisons later
                letters.append(-1)

        # OPT: Pull length calculation out of the loop
        ver_parts1_len = len(ver_parts1)
        ver_parts2_len = len(ver_parts2)

        # Iterate through the components
        for v1, v2 in zip(ver_parts1, ver_parts2):

            # If the string components are equal, the numerical
            # components will be equal too.
            if v1 == v2:
                continue

            # If one of the components begins with a "0" then they
            # are compared as floats so that 1.1 > 1.02; else ints.
            if v1[0] != "0" and v2[0] != "0":
                v1 = int(v1)
                v2 = int(v2)
            else:
                # handle the 0.060 == 0.060 case.
                v1 = v1.rstrip("0")
                v2 = v2.rstrip("0")

            # If they are not equal, the higher value wins.
            c = cmp(v1, v2)
            if c:
                return c

        if ver_parts1_len > ver_parts2_len:
            return 1
        elif ver_parts2_len > ver_parts1_len:
            return -1

        # The dotted components were equal. Let's compare any single
        # letter suffixes.
        if letters[0] != letters[1]:
            return cmp(letters[0], letters[1])

    # The dotted components were equal, so remove them from our lists
    # leaving only suffixes.
    del parts1[0]
    del parts2[0]

    # OPT: Pull length calculation out of the loop
    parts1_len = len(parts1)
    parts2_len = len(parts2)

    # Iterate through the suffixes
    for x in range(max(parts1_len, parts2_len)):

        # If we're at the end of one of our lists, we need to use
        # the next suffix from the other list to decide who wins.
        if x == parts1_len:
            match = suffix_regexp.match(parts2[x])
            val = suffix_value[match.group(1)]
            if val:
                return cmp(0, val)
            return cmp(0, int("0"+match.group(2)))
        if x == parts2_len:
            match = suffix_regexp.match(parts1[x])
            val = suffix_value[match.group(1)]
            if val:
                return cmp(val, 0)
            return cmp(int("0"+match.group(2)), 0)

        # If the string values are equal, no need to parse them.
        # Continue on to the next.
        if parts1[x] == parts2[x]:
            continue

        # Match against our regular expression to make a split between
        # "beta" and "1" in "beta1"
        match1 = suffix_regexp.match(parts1[x])
        match2 = suffix_regexp.match(parts2[x])

        # If our int'ified suffix names are different, use that as the basis
        # for comparison.
        c = cmp(suffix_value[match1.group(1)], suffix_value[match2.group(1)])
        if c:
            return c

        # Otherwise use the digit as the basis for comparison.
        c = cmp(int("0"+match1.group(2)), int("0"+match2.group(2)))
        if c:
            return c

    # Our versions had different strings but ended up being equal.
    # The revision holds the final difference.
    return cmp(rev1, rev2)


class CPV(base.base):
    """base ebuild package class

    :ivar category: str category
    :ivar package: str package
    :ivar key: strkey (cat/pkg)
    :ivar version: str version
    :ivar revision: str revision
    :ivar versioned_atom: atom matching this exact version
    :ivar unversioned_atom: atom matching all versions of this package
    """

    __slots__ = ("cpvstr", "key", "category", "package", "version", "revision", "fullver")

    def __init__(self, *args, versioned=None):
        """
        Can be called with one string or with three string args.

        If called with one arg that is the cpv string. (See :obj:`parser`
        for allowed syntax).

        If called with three args they are the category, package and
        version components of the cpv string respectively.
        """
        for x in args:
            if not isinstance(x, str):
                raise TypeError(f"all args must be strings, got {args!r}")

        l = len(args)
        if l == 1:
            cpvstr = args[0]
            if versioned is None:
                raise TypeError(
                    f"single argument invocation requires versioned kwarg; {cpvstr!r}")
        elif l == 2:
            cpvstr = f"{args[0]}/{args[1]}"
            versioned = False
        elif l == 3:
            cpvstr = f"{args[0]}/{args[1]}-{args[2]}"
            versioned = True
        else:
            raise TypeError(
                f"CPV takes 1 arg (cpvstr), 2 (cat, pkg), or 3 (cat, pkg, ver): got {args!r}")

        try:
            category, pkgver = cpvstr.rsplit("/", 1)
        except ValueError:
            # occurs if the rsplit yields only one item
            raise InvalidCPV(cpvstr, 'no package or version components')
        if not isvalid_cat_re.match(category):
            raise InvalidCPV(cpvstr, 'invalid category name')
        sf = object.__setattr__
        sf(self, 'category', category)
        sf(self, 'cpvstr', cpvstr)
        pkg_chunks = pkgver.split("-")
        lpkg_chunks = len(pkg_chunks)
        if versioned:
            if lpkg_chunks == 1:
                raise InvalidCPV(cpvstr, 'missing package version')
            if isvalid_rev(pkg_chunks[-1]):
                if lpkg_chunks < 3:
                    # needs at least ('pkg', 'ver', 'rev')
                    raise InvalidCPV(
                        cpvstr, 'missing package name, version, and/or revision')
                rev = Revision(pkg_chunks.pop(-1)[1:])
                if rev == 0:
                    # reset stored cpvstr to drop -r0+
                    sf(self, 'cpvstr', f"{category}/{'-'.join(pkg_chunks)}")
                elif rev[0] == '0':
                    # reset stored cpvstr to drop leading zeroes from revision
                    sf(self, 'cpvstr', f"{category}/{'-'.join(pkg_chunks)}-r{int(rev)}")
                sf(self, 'revision', rev)
            else:
                sf(self, 'revision', Revision(''))

            if not isvalid_version_re.match(pkg_chunks[-1]):
                raise InvalidCPV(cpvstr, f"invalid version '{pkg_chunks[-1]}'")
            sf(self, 'version', pkg_chunks.pop(-1))
            if self.revision:
                sf(self, 'fullver', f"{self.version}-r{self.revision}")
            else:
                sf(self, 'fullver', self.version)

            if not isvalid_pkg_name(pkg_chunks):
                raise InvalidCPV(cpvstr, 'invalid package name')
            sf(self, 'package', '-'.join(pkg_chunks))
            sf(self, 'key', f"{category}/{self.package}")
        else:
            if not isvalid_pkg_name(pkg_chunks):
                raise InvalidCPV(cpvstr, 'invalid package name')
            sf(self, 'revision', None)
            sf(self, 'fullver', None)
            sf(self, 'version', None)
            sf(self, 'key', cpvstr)
            sf(self, 'package', '-'.join(pkg_chunks))

    def __hash__(self):
        return hash(self.cpvstr)

    def __repr__(self):
        return '<%s cpvstr=%s @%#8x>' % (
             self.__class__.__name__, getattr(self, 'cpvstr', None), id(self))

    def __str__(self):
        return getattr(self, 'cpvstr', 'None')

    def __eq__(self, other):
        try:
            if self.cpvstr == other.cpvstr:
                return True
            if self.category == other.category and self.package == other.package:
                return ver_cmp(self.version, self.revision, other.version, other.revision) == 0
        except AttributeError:
            pass
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        try:
            if self.category == other.category:
                if self.package == other.package:
                    return ver_cmp(self.version, self.revision, other.version, other.revision) < 0
                return self.package < other.package
            return self.category < other.category
        except AttributeError:
            raise TypeError(
                "'<' not supported between instances of "
                f"{self.__class__.__name__!r} and {other.__class__.__name__!r}"
            )

    def __le__(self, other):
        try:
            if self.category == other.category:
                if self.package == other.package:
                    return ver_cmp(self.version, self.revision, other.version, other.revision) <= 0
                return self.package < other.package
            return self.category < other.category
        except AttributeError:
            raise TypeError(
                "'<=' not supported between instances of "
                f"{self.__class__.__name__!r} and {other.__class__.__name__!r}"
            )

    def __gt__(self, other):
        try:
            if self.category == other.category:
                if self.package == other.package:
                    return ver_cmp(self.version, self.revision, other.version, other.revision) > 0
                return self.package > other.package
            return self.category > other.category
        except AttributeError:
            raise TypeError(
                "'>' not supported between instances of "
                f"{self.__class__.__name__!r} and {other.__class__.__name__!r}"
            )

    def __ge__(self, other):
        try:
            if self.category == other.category:
                if self.package == other.package:
                    return ver_cmp(self.version, self.revision, other.version, other.revision) >= 0
                return self.package > other.package
            return self.category > other.category
        except AttributeError:
            raise TypeError(
                "'>=' not supported between instances of "
                f"{self.__class__.__name__!r} and {other.__class__.__name__!r}"
            )

    @property
    def versioned_atom(self):
        if self.version is not None:
            return atom.atom(f"={self.cpvstr}")
        return self.unversioned_atom

    @property
    def unversioned_atom(self):
        return atom.atom(self.key)

    @classmethod
    def versioned(cls, *args):
        return cls(versioned=True, *args)

    @classmethod
    def unversioned(cls, *args):
        return cls(versioned=False, *args)


class VersionedCPV(CPV):

    __slots__ = ()

    def __init__(self, *args):
        super().__init__(*args, versioned=True)


class UnversionedCPV(CPV):

    __slots__ = ()

    def __init__(self, *args):
        super().__init__(*args, versioned=False)
